% --------------------------------------------------------------------- %
% Define the first example program: quotient-remainder.			%
% --------------------------------------------------------------------- %

g [[ program
        procedure quotient_remainder (var q,r; val x,y);
           pre  0 < y;
           post ^x = q * ^y + r /\ r < ^y;

           r := x;
           q := 0;
           assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
             with r < ^r
           while ~(r < y) do
              r := r - y;
              q := ++q
           od
        end procedure;

        quotient_remainder(q,r;7,3)
     end program
     [ q = 2 /\ r = 1 ]
  ]];;

e(VCG_TAC);;

% 920.1s = 15 min, 20.1 sec %

%
OK..
For procedure `quotient_remainder`,

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r}
       q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "ASSIGN" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y
      {(^x = (q + 1) * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "SEQ" rule, we have
   [[ {(^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\ r - y < ^r}
       r := r - y; q := ++q
      {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ r < ^r} ]]

By the "WHILE" rule, we have
   [[ {^x = q * ^y + r /\ 0 < y /\ ^y = y}
       assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
         with r < ^r
       while ~(r < y) do
          r := r - y; q := ++q
       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + r /\ 0 < y /\ ^y = y}
       q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "ASSIGN" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x
      {^x = 0 * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0
      {^x = q * ^y + r /\ 0 < y /\ ^y = y} ]]

By the "SEQ" rule, we have
   [[ {^x = 0 * ^y + x /\ 0 < y /\ ^y = y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with verification conditions
   "[[[ {((^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(r < y)) /\
         r = ^r ==>
         (^x = (q + 1) * ^y + (r - y) /\ 0 < y /\ ^y = y) /\
         r - y < ^r} ]];
     [[ {(^x = q * ^y + r /\ 0 < y /\ ^y = y) /\ ~(~(r < y)) ==>
         ^x = q * ^y + r /\ r < ^y} ]]]"

By precondition strengthening, we have
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y}
       r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                         with r < ^r
                       while ~(r < y) do
                          r := r - y; q := ++q
                       od
      {^x = q * ^y + r /\ r < ^y} ]]
with additional verification condition
   [[ {(^q = q /\ ^r = r /\ ^x = x /\ ^y = y /\ true) /\ 0 < y ==>
       ^x = 0 * ^y + x /\ 0 < y /\ ^y = y} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure quotient_remainder:

By the call graph progress from procedure quotient_remainder to quotient_remainder, we have
   [[ {0 < y /\ true} quotient_remainder-<>->quotient_remainder {false} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(0 < 3 /\ true) /\
       (!q r x1 y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)}
       quotient_remainder(q,r;7,3)
      {q = 2 /\ r = 1} ]]

By precondition strengthening, we have
   [[ {true} quotient_remainder(q,r;7,3) {q = 2 /\ r = 1} ]]
with additional verification condition
   [[ {true ==> (0 < 3 /\ true) /\
                (!q r x1 y1. 7 = q * 3 + r /\ r < 3 ==> q = 2 /\ r = 1)} ]]

4 subgoals
"0 < 3 /\
 (!q r x1 y1. (7 = (q * 3) + r) /\ r < 3 ==> (q = 2) /\ (r = 1))"

"!^x q ^y r y.
  ((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ r < y ==>
  (^x = (q * ^y) + r) /\ r < ^y"

"!^x q ^y r y ^r.
  (((^x = (q * ^y) + r) /\ 0 < y /\ (^y = y)) /\ ~r < y) /\ (r = ^r) ==>
  ((^x = ((q + 1) * ^y) + (r - y)) /\ 0 < y /\ (^y = y)) /\ (r - y) < ^r"

"!^q q ^r r ^x x ^y y.
  ((^q = q) /\ (^r = r) /\ (^x = x) /\ (^y = y)) /\ 0 < y ==>
  (^x = (0 * ^y) + x) /\ 0 < y /\ (^y = y)"

() : void
Run time: 827.5s
Garbage collection time: 92.6s
Intermediate theorems generated: 53837

%


% 1st subgoal (head of quotient_remainder: %

%e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(REWRITE_TAC[MULT_CLAUSES]);;
e(REWRITE_TAC[ADD_CLAUSES]);;

or%

e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN REWRITE_TAC[MULT_CLAUSES]
  THEN REWRITE_TAC[ADD_CLAUSES]);;

% 2nd subgoal (loop invariant): %

%e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(POP_ASSUM (REWRITE_THM o SYM));;
e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS]));;
e(REWRITE_TAC[SUB_RIGHT_LESS]);;
e(IMP_RES2_THEN REWRITE_THM LESS_LESS_EQ_TRANS);;
e(IMP_RES_THEN (REWRITE_THM o ONCE_REWRITE_RULE[ADD_SYM] o
                  MATCH_MP LESS_ADD_NONZERO o NOT_EQ_SYM)
               LESS_NOT_EQ);;
e(ASM_REWRITE_TAC[]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD1;MULT_CLAUSES]);;
e(REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]);;
e(CONV_TAC ((RAND_CONV o RAND_CONV) (REWR_CONV ADD_SYM)));;
e(IMP_RES_THEN REWRITE_THM SUB_ADD);;%

e(REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN POP_ASSUM (REWRITE_THM o SYM)
  THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS])
  THEN REWRITE_TAC[SUB_RIGHT_LESS]
  THEN IMP_RES2_THEN REWRITE_THM LESS_LESS_EQ_TRANS
  THEN IMP_RES_THEN (REWRITE_THM o ONCE_REWRITE_RULE[ADD_SYM] o
                  MATCH_MP LESS_ADD_NONZERO o NOT_EQ_SYM)
               LESS_NOT_EQ
  THEN ASM_REWRITE_TAC[]
  THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD1;MULT_CLAUSES]
  THEN REWRITE_TAC[(SYM o SPEC_ALL)ADD_ASSOC]
  THEN CONV_TAC ((RAND_CONV o RAND_CONV) (REWR_CONV ADD_SYM))
  THEN IMP_RES_THEN REWRITE_THM SUB_ADD);;

% 3rd subgoal (loop exit): %

%e(REPEAT GEN_TAC);;
e(STRIP_TAC);;
e(ASM_REWRITE_TAC[]);;%

e(REPEAT GEN_TAC
  THEN STRIP_TAC
  THEN ASM_REWRITE_TAC[]);;

% 4th subgoal (main body): %

%e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV));;
e(REWRITE_TAC[]);;
e(REPEAT GEN_TAC);;
e(DISCH_TAC);;
e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (q * 3) + r) /\ r < 3","r:num")));;
e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE));;
e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + r) /\ r < 3","q:num")));;
e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE));;
e(SUBGOAL_THEN "(7 = (2 * 3) + 1) /\ 1 < 3" ASSUME_TAC);;

 e(CONV_TAC (DEPTH_CONV LESS_CONV));;
 e(REWRITE_TAC[num_CONV "7";num_CONV "1"]);;
 e(REWRITE_TAC[ADD_CLAUSES;INV_SUC_EQ]);;
 e(REWRITE_TAC[num_CONV "2"]);;
 e(REWRITE_TAC[MULT_CLAUSES]);;
 e(CONV_TAC (DEPTH_CONV ADD_CONV));;
 e(REWRITE_TAC[]);;

 e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (2 * 3) + r) /\ r < 3","1:num")));;
 e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE));;
 e(FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + 1) /\ 1 < 3","2:num")));;
 e(FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE));;%

e(CONV_TAC (ONCE_DEPTH_CONV LESS_CONV)
  THEN REWRITE_TAC[]
  THEN REPEAT GEN_TAC
  THEN DISCH_TAC
  THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (q * 3) + r) /\ r < 3","r:num"))
  THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE)
  THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + r) /\ r < 3","q:num"))
  THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE)
  THEN SUBGOAL_THEN "(7 = (2 * 3) + 1) /\ 1 < 3" ASSUME_TAC
  THENL
   [ 
     CONV_TAC (DEPTH_CONV LESS_CONV)
     THEN REWRITE_TAC[num_CONV "7";num_CONV "1"]
     THEN REWRITE_TAC[ADD_CLAUSES;INV_SUC_EQ]
     THEN REWRITE_TAC[num_CONV "2"]
     THEN REWRITE_TAC[MULT_CLAUSES]
     THEN CONV_TAC (DEPTH_CONV ADD_CONV)
     THEN REWRITE_TAC[];

     FIRST_ASSUM (ASSUME_TAC o EXISTS("?r. (7 = (2 * 3) + r) /\ r < 3","1:num"))
     THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP DIV_UNIQUE)
     THEN FIRST_ASSUM (ASSUME_TAC o EXISTS("?q. (7 = (q * 3) + 1) /\ 1 < 3","2:num"))
     THEN FIRST_ASSUM (REWRITE_THM o SYM o MATCH_MP MOD_UNIQUE)
  ]);;

let quotient_remainder = top_thm();;
print_vcg_thm quotient_remainder;;
%
|- [[ program
         procedure quotient_remainder(q,r;x,y);
            global ;
            pre  0 < y;
            post ^x = q * ^y + r /\ r < ^y;
            recurses with false;

            r := x; q := 0; assert ^x = q * ^y + r /\ 0 < y /\ ^y = y
                              with r < ^r
                            while ~(r < y) do
                               r := r - y; q := ++q
                            od
         end procedure;

         quotient_remainder(q,r;7,3)
      end program
      [q = 2 /\ r = 1] ]]
%



