% ----------------------------------------------------------------------- %
% Define the sixth example program: Generalized McCarthy's "91" function. %
% ----------------------------------------------------------------------- %

g [[ program
        procedure p91a(var y; val x);
	   global a, b, c, d;
           pre  c = 3  /\  d = (c - 1)*b + 1 ;
           post (^a = a /\ ^b = b /\ ^c = c /\ ^d = d /\ ^x = x) /\
                a < x  =>  y = x - b  |  y = a + 1 - b;
           calls p91a with a + 1 - x < a + 1 - ^x;
           recurses  with a + 1 - x < ^z;

           if a < x then y := x - b
           else
	      y := x + d;
              p91a(y; y);
	      p91a(y; y);
              p91a(y; y)
           fi
        end procedure;

        a := 100;
	b := 10;
	c := 3;
	d := 21;
        p91a(y; 77)

     end program
     [ y = 91 ]
  ]];;

e(VCG_TAC);;

% 1401.2s = 23 min, 21.2 sec %

%<
#e(VCG_TAC);;
OK..
For procedure `p91`,

By the "ASSIGN" rule, we have
   [[ {(100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)}
       x := y - 10
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "CALL" rule, we have
   [[ {(true /\ 101 - x < 101 - ^y) /\
       (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                (100 < ^y => x1 = ^y - 10 | x1 = 91))}
       p91(x;x)
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "CALL" rule, we have
   [[ {(true /\ 101 - (y + 11) < 101 - ^y) /\
       (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91)))}
       p91(x;y + 11)
      {(true /\ 101 - x < 101 - ^y) /\
       (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                (100 < ^y => x1 = ^y - 10 | x1 = 91))} ]]

By the "SEQ" rule, we have
   [[ {(true /\ 101 - (y + 11) < 101 - ^y) /\
       (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91)))}
       p91(x;y + 11); p91(x;x)
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By the "IF" rule, we have
   [[ {(100 < y => (100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)
           | (true /\ 101 - (y + 11) < 101 - ^y) /\
             (!x y1. (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
                     (true /\ 101 - x < 101 - ^y) /\
                     (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                              (100 < ^y => x1 = ^y - 10 | x1 = 91))))}
       if 100 < y then x := y - 10 else p91(x;y + 11); p91(x;x) fi
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]

By precondition strengthening, we have
   [[ {(^x = x /\ ^y = y /\ true) /\ true}
       if 100 < y then x := y - 10 else p91(x;y + 11); p91(x;x) fi
      {(100 < ^y => x = ^y - 10 | x = 91)} ]]
with additional verification condition
   [[ {(^x = x /\ ^y = y /\ true) /\ true ==>
       (100 < y => (100 < ^y => y - 10 = ^y - 10 | y - 10 = 91)
           | (true /\ 101 - (y + 11) < 101 - ^y) /\
             (!x y1.
               (100 < y + 11 => x = (y + 11) - 10 | x = 91) ==>
               (true /\ 101 - x < 101 - ^y) /\
               (!x1 y1. (100 < x => x1 = x - 10 | x1 = 91) ==>
                        (100 < ^y => x1 = ^y - 10 | x1 = 91))))} ]]

Examining the structure of the procedure call graph:

Traversing the call graph back from the procedure p91:

By the call graph progress from procedure p91 to p91, we have
   [[ {true /\ (!x y1. 101 - y1 < 101 - y ==> 101 - y1 < ^z)}
      p91-<>->p91
      {101 - y < ^z} ]]

Generating the undiverted recursion verification condition
   [[ {true /\ 101 - y = ^z ==>
       (!x y1. 101 - y1 < 101 - y ==> 101 - y1 < ^z)} ]]

For the main body,

By the "CALL" rule, we have
   [[ {(true /\ true) /\
       (!a y1. (100 < 77 => a = 77 - 10 | a = 91) ==> a = 91)}
       p91(a;77)
      {a = 91} ]]

By precondition strengthening, we have
   [[ {true} p91(a;77) {a = 91} ]]
with additional verification condition
   [[ {true ==> (true /\ true) /\
                (!a y1. (100 < 77 => a = 77 - 10 | a = 91) ==> a = 91)} ]]

3 subgoals
"!a y1. (100 < 77 => (a = 77 - 10) | (a = 91)) ==> (a = 91)"

"!y ^z.
  (101 - y = ^z) ==> (!x y1. (101 - y1) < (101 - y) ==> (101 - y1) < ^z)"

"!^x x ^y y.
  (^x = x) /\ (^y = y) ==>
  (100 < y => 
   (100 < ^y => (y - 10 = ^y - 10) | (y - 10 = 91)) | 
   ((101 - (y + 11)) < (101 - ^y) /\
    (!x' y1.
      (100 < (y + 11) => (x' = (y + 11) - 10) | (x' = 91)) ==>
      (101 - x') < (101 - ^y) /\
      (!x1 y1'.
        (100 < x' => (x1 = x' - 10) | (x1 = 91)) ==>
        (100 < ^y => (x1 = ^y - 10) | (x1 = 91))))))"

() : void
Run time: 1280.3s
Garbage collection time: 120.9s
Intermediate theorems generated: 76317

>%



let SUB_ADD_LESS_SUB =
 TAC_PROOF
  (([],
    "!a b c. b < a /\ 0 < c ==> (a - (b + c)) < a - b"),
   REPEAT STRIP_TAC
   THEN REWRITE_TAC[SUB_PLUS]
   THEN MP_TAC (SPECL["a - b";"c:num"] SUB_LESS_EQ)
   THEN REWRITE_TAC[LESS_OR_EQ]
   THEN REWRITE_TAC[SUB_EQ_EQ_0]
   THEN REWRITE_TAC[SUB_EQ_0]
   THEN IMP_RES_THEN (REWRITE_THM o NOT_EQ_SYM) LESS_NOT_EQ
   THEN ASM_REWRITE_TAC[(SYM o SPEC_ALL)NOT_LESS]
  );;

%
g "!a b c. b < a /\ 0 < c ==> (a - (b + c)) < a - b";;
e(REPEAT STRIP_TAC);;
e(REWRITE_TAC[SUB_PLUS]);;
e(MP_TAC (SPECL["a - b";"c:num"] SUB_LESS_EQ));;
e(REWRITE_TAC[LESS_OR_EQ]);;
e(REWRITE_TAC[SUB_EQ_EQ_0]);;
e(REWRITE_TAC[SUB_EQ_0]);;
e(IMP_RES_THEN (REWRITE_THM o NOT_EQ_SYM) LESS_NOT_EQ);;
e(ASM_REWRITE_TAC[(SYM o SPEC_ALL)NOT_LESS]);;
%


% There are four cases to consider:
     y < 90, 90 <= y < 100, y = 100, and y > 100.
%

%
e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
 e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS]));;
 e(USE_THEN REWRITE_THM SUB_ADD_LESS_SUB);;

  e(CONV_TAC (DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[]);;
  e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
  e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;

  e(GEN_TAC);;
  e(COND_CASES_TAC);;

   e(POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11");
                                          LESS_MONO_ADD_EQ]));;
   e(SUBGOAL_THEN "10 <= 11" (REWRITE_THM o MATCH_MP LESS_EQ_ADD_SUB));;

    e(REWRITE_TAC[LESS_OR_EQ]);;
    e(CONV_TAC (DEPTH_CONV LESS_CONV));;
    e(REWRITE_TAC[]);;

    e(CONV_TAC (DEPTH_CONV SUBTRACT_CONV));;
    e(DISCH_THEN REWRITE_THM);;
    e(USE_THEN REWRITE_THM SUB_ADD_LESS_SUB);;

     e(CONV_TAC (DEPTH_CONV LESS_CONV));;
     e(REWRITE_TAC[]);;
     e(CONV_TAC (ONCE_DEPTH_CONV num_CONV));;
     e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;

     e(GEN_TAC);;
     e(UNDISCH_LAST_TAC);;
     e(POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[LESS_OR_EQ]));;

      e(IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[ADD1]) LESS_SUC_NOT);;

      e(POP_ASSUM REWRITE_THM);;
      e(CONV_TAC (DEPTH_CONV ADD_CONV));;
      e(CONV_TAC (DEPTH_CONV LESS_CONV));;
      e(CONV_TAC (DEPTH_CONV SUBTRACT_CONV));;
      e(REWRITE_TAC[]);;

  e(UNDISCH_LAST_TAC);;
  e(REWRITE_TAC[NOT_LESS]);;
  e(DISCH_THEN (MP_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11")]));;
  e(REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]);;
  e(DISCH_TAC);;
  e(DISCH_THEN REWRITE_THM);;
  e(CONV_TAC (DEPTH_CONV LESS_CONV));;
  e(REWRITE_TAC[]);;
  e(SUBGOAL_THEN "91 = y + (91 - y)" ONCE_REWRITE_THM);;

   e(ONCE_REWRITE_TAC[ADD_SYM]);;
   e(USE_IMP_THEN REWRITE_THM SUB_ADD);;
   e(IMP_RES_TAC LESS_EQ_TRANS);;
   e(POP_TAC THEN POP_TAC THEN POP_TAC);;
   e(POP_ASSUM MATCH_MP_TAC);;
   e(ONCE_REWRITE_TAC[LESS_OR_EQ]);;
   e(CONV_TAC (DEPTH_CONV LESS_CONV));;
   e(REWRITE_TAC[]);;

   e(MATCH_MP_TAC SUB_ADD_LESS_SUB);;
   e(CONV_TAC (RATOR_CONV (ONCE_DEPTH_CONV num_CONV)));;
   e(IMP_RES_TAC LESS_EQ_IMP_LESS_SUC);;
   e(POP_ASSUM REWRITE_THM);;
   e(REWRITE_TAC[(SYM o SPEC_ALL)SUB_LESS_0]);;
   e(IMP_RES_TAC LESS_TRANS);;
   e(FIRST_ASSUM MATCH_MP_TAC);;
   e(REWRITE_TAC[LESS_EQ]);;
   e(REWRITE_TAC[LESS_OR_EQ]);;
   e(DISJ2_TAC);;
   e(REWRITE_TAC(map (SYM o num_CONV) ["90";"91"]));;

e(REPEAT GEN_TAC);;
e(DISCH_THEN REWRITE_THM);;

e(CONV_TAC (DEPTH_CONV LESS_CONV));;
e(REWRITE_TAC[]);;
%



%
e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM
  THEN COND_CASES_TAC
  THENL
   [ REWRITE_TAC[];

     REWRITE_TAC[]
     THEN POP_ASSUM (ASSUME_TAC o REWRITE_RULE[NOT_LESS])
     THEN USE_THEN REWRITE_THM SUB_ADD_LESS_SUB
     THENL
      [ CONV_TAC (DEPTH_CONV LESS_CONV)
        THEN REWRITE_TAC[]
        THEN CONV_TAC (ONCE_DEPTH_CONV num_CONV)
        THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC;

        GEN_TAC
        THEN COND_CASES_TAC
        THENL
         [ POP_ASSUM (ASSUME_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11");
                                          LESS_MONO_ADD_EQ])
           THEN SUBGOAL_THEN "10 <= 11" (REWRITE_THM o MATCH_MP LESS_EQ_ADD_SUB)
           THENL
            [ REWRITE_TAC[LESS_OR_EQ]
              THEN CONV_TAC (DEPTH_CONV LESS_CONV)
              THEN REWRITE_TAC[];

              CONV_TAC (DEPTH_CONV SUBTRACT_CONV)
              THEN DISCH_THEN REWRITE_THM
              THEN USE_THEN REWRITE_THM SUB_ADD_LESS_SUB
              THENL
               [ CONV_TAC (DEPTH_CONV LESS_CONV)
                 THEN REWRITE_TAC[]
                 THEN CONV_TAC (ONCE_DEPTH_CONV num_CONV)
                 THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC;

                 GEN_TAC
                 THEN UNDISCH_LAST_TAC
                 THEN POP_ASSUM (STRIP_ASSUME_TAC o REWRITE_RULE[LESS_OR_EQ])
                 THENL
                  [ IMP_RES_THEN (REWRITE_THM o REWRITE_RULE[ADD1]) LESS_SUC_NOT;

                    POP_ASSUM REWRITE_THM
                    THEN CONV_TAC (DEPTH_CONV ADD_CONV)
                    THEN CONV_TAC (DEPTH_CONV LESS_CONV)
                    THEN CONV_TAC (DEPTH_CONV SUBTRACT_CONV)
                    THEN REWRITE_TAC[]
                  ]
               ]
            ];

           UNDISCH_LAST_TAC
           THEN REWRITE_TAC[NOT_LESS]
           THEN DISCH_THEN (MP_TAC o REWRITE_RULE[SYM(ADD_CONV "89 + 11")])
           THEN REWRITE_TAC[LESS_EQ_MONO_ADD_EQ]
           THEN DISCH_TAC
           THEN DISCH_THEN REWRITE_THM
           THEN CONV_TAC (DEPTH_CONV LESS_CONV)
           THEN REWRITE_TAC[]
           THEN SUBGOAL_THEN "91 = y + (91 - y)" ONCE_REWRITE_THM
           THENL
            [ ONCE_REWRITE_TAC[ADD_SYM]
              THEN USE_IMP_THEN REWRITE_THM SUB_ADD
              THEN IMP_RES_TAC LESS_EQ_TRANS
              THEN POP_TAC THEN POP_TAC THEN POP_TAC
              THEN POP_ASSUM MATCH_MP_TAC
              THEN ONCE_REWRITE_TAC[LESS_OR_EQ]
              THEN CONV_TAC (DEPTH_CONV LESS_CONV)
              THEN REWRITE_TAC[];

              MATCH_MP_TAC SUB_ADD_LESS_SUB
              THEN CONV_TAC (RATOR_CONV (ONCE_DEPTH_CONV num_CONV))
              THEN IMP_RES_TAC LESS_EQ_IMP_LESS_SUC
              THEN POP_ASSUM REWRITE_THM
              THEN REWRITE_TAC[(SYM o SPEC_ALL)SUB_LESS_0]
              THEN IMP_RES_TAC LESS_TRANS
              THEN FIRST_ASSUM MATCH_MP_TAC
              THEN REWRITE_TAC[LESS_EQ]
              THEN REWRITE_TAC[LESS_OR_EQ]
              THEN DISJ2_TAC
              THEN REWRITE_TAC(map (SYM o num_CONV) ["90";"91"])
            ]
         ]
      ]
   ]);;
%

%
e(REPEAT GEN_TAC
  THEN DISCH_THEN REWRITE_THM);;

e(CONV_TAC (DEPTH_CONV LESS_CONV)
  THEN REWRITE_TAC[]);;
  %



%
let p91 = top_thm();;
print_vcg_thm p91;;
%

%
|- [[ program procedure p91(x;y);
                 global ;
                 pre  true;
                 post (100 < ^y => x = ^y - 10 | x = 91);
                 calls p91 with 101 - y < 101 - ^y;
                 recurses with 101 - y < ^z;

                 if 100 < y
                    then x := y - 10
                    else p91(x;y + 11); p91(x;x)
                 fi
              end procedure; p91(a;77) end program
      [a = 91] ]]
%



