% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the syntax of the assertion language.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `assert_syntax`;;

new_parent `variables`;;


loadf `semantics_tactics`;;


%===========================================================%
% Assertions, etc. are written in a language of their own.  %
% Here is where we define the syntax of that language.      %
%===========================================================%

%===========================================================%
% First we define the datatypes for assertion expressions.  %
% vexp is a value-yielding expression, of type num.         %
% aexp is an assertion expression, yielding a boolean.      %
%===========================================================%


let vexp_Axiom,vexp_induct,vexp_cases,vexp_one_one,vexp_distinct =
    new_rec_type
      `vexp`
      `  ANUM num | AVAR var
       | APLUS vexp vexp | AMINUS vexp vexp | AMULT vexp vexp`;;


let aexp_Axiom,aexp_induct,aexp_cases,aexp_one_one,aexp_distinct =
    new_rec_type
      `aexp`
      `  ATRUE | AFALSE
       | AEQ vexp vexp | ALESS vexp vexp
       | ALLESS (vexp list) (vexp list)
       | AAND aexp aexp | AOR aexp aexp | ANOT aexp
       | AIMP aexp aexp | AEQB aexp aexp
       | ACOND aexp aexp aexp
       | ACLOSE aexp
       | AFORALL var aexp | AEXISTS var aexp`;;

%===========================================================%
%  ANUM is a number; AVAR is a program variable; APLUS and  %
%  AMINUS are the assertion plus and minus operators.       %
%  (Note that 0 - x = x here.)                              %
%  AMULT is the corrsponding multiplication operator.       %
%  AEQ is the relation of equality of numbers.              %
%  ALESS is the relation of less-than on numbers.           %
%  ATRUE and AFALSE are the assertion boolean constants.    %
%  AAND, AOR, and ANOT are the assertion boolean operators. %
%  AIMP is implication, and AEQB is equality of booleans.   %
%  ACOND is a boolean conditional expression (a => b | c).  %
%  ACLOSE is 'closed', means the arg is true in all states. %
%  AFORALL and AEXISTS are the universal and existential    %
%  quantifiers.                                             %
%===========================================================%


close_theory();;

