% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definitions of the syntax of the programming
  language.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `prog_syntax`;;

new_parent `assert_syntax`;;


loadf `semantics_tactics`;;


%================================================================%
% SYNTAX:                                                        %
% Here we define the datatypes for programming language          %
% expressions and commands.  Note that these are recursive.      %
%================================================================%


% Variable lists:                                 %
%   We define 'vars' as a list of identifiers.    %

new_type_abbrev(`vars`, ":var list");;


let exp_Axiom,exp_induct,exp_cases,exp_one_one,exp_distinct =
    new_rec_type
      `exp`
      `NUM num | PVAR var | INC var |
       PLUS exp exp | MINUS exp exp | MULT exp exp`;;


% Expression lists:                               %
%   We define 'exps' as a list of identifiers.    %

new_type_abbrev(`exps`, ":exp list");;


let bexp_Axiom,bexp_induct,bexp_cases,bexp_one_one,bexp_distinct =
    new_rec_type
      `bexp`
      `  EQ exp exp | LESS exp exp | LLESS exps exps
       | AND bexp bexp | OR bexp bexp | NOT bexp`;;


let cmd_Axiom,cmd_induct,cmd_cases,cmd_one_one,cmd_distinct =
    new_rec_type
      `cmd`
      `  SKIP
       | ABORT
       | ASSIGN var exp
       | SEQ cmd cmd
       | IF bexp cmd cmd
       | WHILE aexp aexp bexp cmd
       | CALL string vars exps `;;


% Progress environments:                               %
%   We define ':prog_env' as a function from strings   %
%   (procedure names) to aexp (progress expressions).  %

new_type_abbrev(`prog_env`, ":string -> aexp");;


let decl_Axiom,decl_induct,decl_cases,decl_one_one,decl_distinct =
    new_rec_type
      `decl`
      `  PROC string vars vars vars aexp aexp prog_env aexp cmd
       | DSEQ decl decl
       | DEMPTY `;;


let prog_Axiom,prog_induct,prog_cases,prog_one_one,prog_distinct =
    new_rec_type
      `prog`
      `  PROG decl cmd `;;


close_theory();;
