% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% An initial version of a semantics for a small language;
  First given in a structural operational style;
  then expressed as an axiomatic semantics, which is
  proven from the structural operational semantics,
  then the weakest derivable precondition functions
  are defined and proven sound.

  This file contains the definition of the basic Hoare-style predicates.

  By Peter V. Homeier, November 4, 1991.
%

set_flag(`timing`,true);;

set_search_path(  (library_pathname() ^ `/finite_sets/`)
                . (library_pathname() ^ `/string/`)
                . search_path());;

% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `specifications`;;

new_parent `translations`;;


loadf `semantics_tactics`;;


% =================================================================== %
% Define the basic predicates that specify Hoare-axiom relationships. %
% =================================================================== %


% {p} (p is true, in any state) is represented by CLOSE(p) %

let CLOSE =
  new_def
  `CLOSE`
  "CLOSE(p) = !(s:state). A p s";;


% {p}e{q} is represented by ESPEC(p,e,q) %

let ESPEC = new_definition
 (`ESPEC`,
  "ESPEC(p,e,q) = !(s1 s2:state) (n:num).
            A p s1 /\ E e s1 n s2 ==> A q s2");;


% {p}es{q} is represented by ES_SPEC(p,es,q) %

let ES_SPEC = new_definition
 (`ES_SPEC`,
  "ES_SPEC(p,es,q) = !(s1 s2:state) (ns:(num)list).
            A p s1 /\ ES es s1 ns s2 ==> A q s2");;


% {p}b{q} is represented by BSPEC(p,b,q) %

let BSPEC = new_definition
 (`BSPEC`,
  "BSPEC(p,b,q) = !(s1 s2:state) (t:bool). 
            A p s1 /\ B b s1 t s2 ==> A q s2");;


% {p}c{q}/d is represented by CSPEC(p,c,q,d) %

let CSPEC = new_definition
 (`CSPEC`,
  "CSPEC(p,c,q,d) = !(s1 s2:state). A p s1 /\ C c d s1 s2 ==> A q s2");;


% pi{q} is represented by PSPEC(pi,q)   (pi is a program) %

let PSPEC = new_definition
 (`PSPEC`,
  "PSPEC(pi,q) = !(s1:state). P pi s1 ==> A q s1");;



close_theory ();;

