% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% The Sunrise system consists of a deep embedding of the Sunrise
  programming language with its associated assertion language in HOL,
  with their semantics.  Then an axiomatic semantics of the programming
  language is defined and proven correct with respect to the foundational
  semantics.  A Verification Condition Generator function for total
  correctness is defined and is proven to be sound, in that the truth of
  the verification conditions generated suffices to ensure the total
  correctness of the program.

  This file contains the definition of the structure of variables for both
  the assertion language and the programming language.

  By Peter V. Homeier, May 31, 1995.
  Version 1.0 (first public version)
%


% --------------------------------------------------------------------- %
% Create the theory.							%
% --------------------------------------------------------------------- %
new_theory `variables`;;


% --------------------------------------------------------------------- %
% Need the string library.                                              %
% --------------------------------------------------------------------- %
load_library `string`;;


loadf `semantics_tactics`;;


%===========================================================%
% The actual "names" of variables will be defined as a      %
% composite type, containing not only a string but also a   %
% "variant" number; initially this will be zero, but when   %
% this is positive, it indicates a numeric postfix to the   %
% variable name.  This is to make it easier to define       %
% recognizable variants of variable names.  A zero postfix  %
% will not be read; thus (`x`,0) => "x" but (`y`,2) => "y2".%
%===========================================================%

let var_Axiom,var_induct,var_cases,var_one_one,var_distinct =
    new_rec_type
      `var`
      `VAR string num`;;




% =============================================================== %
% We define Base v as the string which is the base of v.          %
% =============================================================== %

let Base = 
    new_recursive_definition false var_Axiom `Base`
      "(Base (VAR s n) = s)";;

% =============================================================== %
% We define Index v as the "variant" number attached to the base. %
% =============================================================== %

let Index = 
    new_recursive_definition false var_Axiom `Index`
      "(Index (VAR s n) = n)";;


let Base_Index = prove_thm
   (`Base_Index`,
    "!x. VAR (Base x) (Index x) = x",
    INDUCT_THEN var_induct ASSUME_TAC  THEN
    REWRITE_TAC[Base;Index]
   );;

%
g "!x. VAR (Base x) (Index x) = x";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[Base;Index]);;
%

let VAR_EQ = prove_thm
   (`VAR_EQ`,
    "!x y. (Base x = Base y) /\ (Index x = Index y) ==> (x = y)",
    INDUCT_THEN var_induct ASSUME_TAC  THEN
    GEN_TAC THEN GEN_TAC  THEN
    INDUCT_THEN var_induct ASSUME_TAC  THEN
    REWRITE_TAC[Base;Index]  THEN
    REPEAT STRIP_TAC  THEN
    ASM_REWRITE_TAC[var_one_one]
   );;

%
g "!x y. (Base x = Base y) /\ (Index x = Index y) ==> (x = y)";;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(GEN_TAC THEN GEN_TAC);;
e(INDUCT_THEN var_induct ASSUME_TAC);;
e(REWRITE_TAC[Base;Index]);;
e(REPEAT STRIP_TAC);;
e(ASM_REWRITE_TAC[var_one_one]);;
%


close_theory();;

