% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% ====================================================================== %
% This file defines the function `prove_translation_thm`,  which derives %
% the theorem relating  the tuples in the  programming language relation %
% with the results of applying the translation functions.   The proof is %
% quite regular,  even for different expression types.   The translation %
% function derives the required translation theorem automatically.       %
%           This is essentially a compiler correctness proof.            %
% ====================================================================== %


let prove_translation_thm lemma_name name source dest defin defin_comps
                          trans_body exists_tac ex_thms =
    %        "E e s1 n s2 =    (n = V (VE e) s1)    /\
                               (s2 = s1 <|s (VE_state e))"                %
    let reln,trans_eqs = dest_eq trans_body in
    %   "E e s1 n s2" , "(n = V (VE e) s1)  /\  (s2 = s1 <|s (VE_state e))"   %
    let rel,quant_vars = strip_comb reln
    %   "E" , ["e"; "s1"; "n"; "s2"]   %
    in
    let source_def = definition `prog_semantics` (source ^ `_DEF`) in
    %   E_DEF   %
    let dest_def = definition `assert_semantics` (dest   ^ `_DEF`)
    %   V_DEF   %
    in
    let rules,ind = (let rules,ind = CONJ_PAIR source_def
                     in CONJUNCTS rules,ind)
    in
    letrec strip_conj tm = if is_conj tm then let c1,c2 = dest_conj tm in
                                              (c1 . strip_conj c2)
                                         else [tm]
    in
    let eqs_terms = map ((\a,b. b,a) o dest_eq) (strip_conj trans_eqs) in
    %   [("V (VE e) s1","n"); ("s1 <|s (VE_state e)","s2")]   %
    let sub_vars = snd (split eqs_terms) in
    %   ["n"; "s2"]   %
    let exists_vars = subtract quant_vars sub_vars in
    %   ["e"; "s1"]   %
    let source_var = hd exists_vars in
    %   "e:exp"   %
    let type_name = (fst o dest_type o snd o dest_var) source_var in
    %   `exp`   %
    let type_induct = theorem `prog_syntax` (type_name ^ `_induct`)
    %   exp_induct   %
    in
    let [V;VE1] = map (fst o strip_comb o fst o dest_eq o snd o strip_forall
                           o concl o hd o CONJUNCTS) [dest_def; defin] in
    let [VE;VE_state] = map (fst o strip_comb o fst o dest_eq o snd
                                 o strip_forall o concl) defin_comps in
    let lemma_term = 
           "!^source_var s ss.
            (       ^V (FST (^VE1 ^source_var ss)) s
             =      ^V (^VE ^source_var) (s <|s ss) )
         /\   ( (s <|s (SND (^VE1 ^source_var ss)))
             = ((s <|s ss) <|s (^VE_state ^source_var)))"
    in
    print_string `Proving lemma  `;print_term lemma_term; print_newline();
    let lemma = PROVE (lemma_term,
                        INDUCT_THEN type_induct STRIP_ASSUME_TAC  THEN
                        REWRITE_TAC (defin . defin_comps)  THEN
                        TRANS_CONT_TAC  THEN
                        ASM_REWRITE_TAC (ex_thms @ [dest_def;s_subst_BND;
                                         V_IDENT_subst;s_subst_IDENT]))
    in
    save_thm (lemma_name, lemma);
    let unique_term = list_mk_forall(quant_vars, mk_imp(reln,trans_eqs)) in
    print_string `Proving unique `;print_term unique_term; print_newline();
    let SEM_INDUCT_TAC = RULE_INDUCT_THEN ind
                         STRIP_ASSUME_TAC STRIP_ASSUME_TAC in
    let unique = PROVE (unique_term,
                        SEM_INDUCT_TAC  THEN
                        REWRITE_ALL_TAC (defin . defin_comps)  THEN
                        TRANS_CONT_TAC  THEN
                        REWRITE_ALL_TAC [dest_def;s_subst_BND;
                                         IDENT_SS_THM;s_subst_IDENT]  THEN
                        ASM_REWRITE_ALL_TAC (ex_thms @ [lemma;s_subst_IDENT]))
    in
    let exists_term = list_mk_forall(exists_vars, subst eqs_terms reln) in
    print_string `Proving exists `;print_term exists_term; print_newline();
    let exists = PROVE (exists_term,
                        INDUCT_THEN type_induct ASSUME_TAC  THEN
                        REPEAT GEN_TAC  THEN
                        REWRITE_TAC (defin . defin_comps)  THEN
                        TRANS_CONT_TAC  THEN
                        REWRITE_TAC [dest_def;s_subst_BND;
                                         IDENT_SS_THM;s_subst_IDENT]
                        THENL exists_tac
                        THEN ASM_REWRITE_TAC (ex_thms @ [lemma;s_subst_IDENT]))
    in
    print_newline();
    let trans_term = list_mk_forall(quant_vars, trans_body) in
    let trans_thm = PROVE (trans_term,
                        REPEAT GEN_TAC  THEN
                        EQ_TAC  THENL
                           [ REWRITE_TAC[unique];
                             STRIP_TAC  THEN
                             ASM_REWRITE_TAC[exists]
                           ])
    in save_thm (name, trans_thm);
       lemma,trans_thm;;


%
let source = `E`;;
let dest   = `V`;;
let defin  = VE_DEF;;
let trans_term  = "!e s1 n s2. E e s1 n s2 =
                  (n = V(VE e) s1)    /\
                  (s2 = s1)";;
let exists_tac    =
         [
            NUM_TAC;
            VAR_TAC;
            PLUS_TAC
       THEN EXISTS_TAC "s1:state"
         ];;


    let defin = new_rec_def (dest ^ source ^ `_DEF`) source_Axiom def_term;;

    let quant_vars,trans_body = strip_forall trans_term;;
    let rel,trans_eqs = dest_eq trans_body;;

    let unique_term = list_mk_forall(quant_vars, mk_imp(rel,trans_eqs));;
    let source_def = definition `prog_semantics` (source ^ `_DEF`);;
    let rules,ind = (let rules,ind = CONJ_PAIR source_def
                     in CONJUNCTS rules,ind);;
    let dest_def = definition `assert_semantics` (dest   ^ `_DEF`);;

    let SEM_INDUCT_TAC = RULE_INDUCT_THEN ind
                         STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
    let unique = PROVE (unique_term,
                        SEM_INDUCT_TAC THEN
                        ASM_REWRITE_TAC[defin;dest_def]);;

    letrec strip_conj tm = if is_conj tm then let c1,c2 = dest_conj tm in
                                              (c1 . strip_conj c2)
                                         else [tm];;

    let eqs_terms = map ((\a,b. b,a) o dest_eq) (strip_conj trans_eqs);;
    let sub_vars = snd (split eqs_terms);;
    let exists_vars = subtract quant_vars sub_vars;;
    let exists_term = list_mk_forall(exists_vars, subst eqs_terms rel);;
    let type_name = (fst o dest_type o snd o dest_var o hd) exists_vars;;
    let type_induct = theorem `prog_syntax` (type_name ^ `_induct`);;
    let exists = PROVE (exists_term,
                        INDUCT_THEN type_induct ASSUME_TAC  THEN
                        REPEAT GEN_TAC  THEN
                        REWRITE_TAC[defin;dest_def]
                        THENL exists_tac
                        THEN ASM_REWRITE_TAC[]);;

    let trans_thm = PROVE (trans_term,
                        REPEAT GEN_TAC  THEN
                        EQ_TAC  THENL
                           [ REWRITE_TAC[unique];
                             STRIP_TAC  THEN
                             ASM_REWRITE_TAC[exists]
                           ]);;

    defin,trans_thm;;
%