
% --------------------------------------------------------------------- %
% Load the programming language semantics rules.                        %
% --------------------------------------------------------------------- %

let E_rules,E_ind = let rules,ind = CONJ_PAIR E_DEF in CONJUNCTS rules,ind;;
let B_rules,B_ind = let rules,ind = CONJ_PAIR B_DEF in CONJUNCTS rules,ind;;
let C_rules,C_ind = let rules,ind = CONJ_PAIR C_DEF in CONJUNCTS rules,ind;;
let D_rules,D_ind = let rules,ind = CONJ_PAIR D_DEF in CONJUNCTS rules,ind;;
let P_rules,P_ind = let rules,ind = CONJ_PAIR P_DEF in CONJUNCTS rules,ind;;


% --------------------------------------------------------------------- %
% Tactics for the rules.						%
% --------------------------------------------------------------------- %

let EXP_TAC = FIRST (map RULE_TAC E_rules);;
let [NUM_TAC;PVAR_TAC;INC_TAC;
     PLUS_TAC;MINUS_TAC;MULT_TAC] = map RULE_TAC E_rules;;

let EXPS_TAC = LIST_INDUCT_TAC;;

let BEXP_TAC = FIRST (map RULE_TAC B_rules);;
let [BEQ_TAC;LESS_TAC;LLESS_TAC;AND_TAC;OR_TAC;NOT_TAC] = map RULE_TAC B_rules;;

let CMD_TAC = FIRST (map RULE_TAC C_rules);;
let [SKIP_TAC;ASSIGN_TAC;SEQ_TAC;IF_T_TAC;IF_F_TAC;
     WHILE_T_TAC;WHILE_F_TAC;CALL_TAC] = map RULE_TAC C_rules;;

let DECL_TAC = FIRST (map RULE_TAC D_rules);;
let [PROC_TAC;DSEQ_TAC;DEMPTY_TAC] = map RULE_TAC D_rules;;

let [PROG_TAC] = map RULE_TAC P_rules;;

% --------------------------------------------------------------------- %
% Rule induction tactics.                                               %
% --------------------------------------------------------------------- %

let  EXP_INDUCT_TAC = RULE_INDUCT_THEN E_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let BEXP_INDUCT_TAC = RULE_INDUCT_THEN B_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let  CMD_INDUCT_TAC = RULE_INDUCT_THEN C_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let DECL_INDUCT_TAC = RULE_INDUCT_THEN D_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;
let PROG_INDUCT_TAC = RULE_INDUCT_THEN P_ind STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;


let mk_str_ind_tac str_ind_thm = RULE_INDUCT_THEN str_ind_thm
                                 STRIP_ASSUME_TAC STRIP_ASSUME_TAC;;

let  EXP_STR_INDUCT_TAC = mk_str_ind_tac E_str_induct;;
let BEXP_STR_INDUCT_TAC = mk_str_ind_tac B_str_induct;;
let  CMD_STR_INDUCT_TAC = mk_str_ind_tac C_str_induct;;
let DECL_STR_INDUCT_TAC = mk_str_ind_tac D_str_induct;;
let PROG_STR_INDUCT_TAC = mk_str_ind_tac P_str_induct;;



let SEMAN_STRIP_ASSUME_TAC =
    %CONJUNCTS_THEN2 STRIP_ASSUME_TAC%
    ((REPEAT_TCL (FIRST_TCL [DISJ_CASES_THEN;CHOOSE_THEN] ))
     (\th. 
       %print_string `Here are they:`; print_newline();%
       (let c1,c2 = CONJ_PAIR th in
       (%print_thm c1; print_newline();
       print_thm c2; print_newline();%
       let c3 = CONJUNCTS c1 in
       (%map print_thm c3; print_newline();%
       STRIP_ASSUME_TAC ((REWRITE_RULE (map SYM c3) c2)
                         ? c2)))) ? STRIP_ASSUME_TAC th));;


% Use as: e(RULE_STRIP_ASSUM_TAC C_rewrite);; %

let RULE_STRIP_ASSUM_TAC rule =
    POP_ASSUM_LIST (\asl. MAP_EVERY SEMAN_STRIP_ASSUME_TAC (rev (map rule asl)));;
