% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

let pthm th =
    print_thm th;
    print_newline();
    th;;

let pairf f g (x,y) = (f x,g y);;
let test f x = f x; x;;
let try f x = f x ? x;;

map new_special_symbol [`<|s`;`<|v`;`<|vs`;`<|a`;
                        `<|xs`;`<|e`;`<|es`;`<|b`;`<|c`;`<|env`;`<|g`;
                        `<|vv`;`<|vsv`;`<|av`;`:==`;`//v`;`<<`]; ();;

loadf `new_type`;;

let ONE_ONE_DEF = definition `bool` `ONE_ONE_DEF`;;
let ONTO_DEF    = definition `bool` `ONTO_DEF`;;

% These operators transform implications to collect antecedents into
  a single conjunction, or else to reverse this and spread them into
  a sequence of individual implications.  Each has type thm->thm.
%

let AND2IMP = REWRITE_RULE[(SYM o SPEC_ALL)AND_IMP_INTRO];;
let IMP2AND = REWRITE_RULE[AND_IMP_INTRO];;


% Here are some useful tactics, that are not included in the standard HOL: %

let PRINT_TAC :tactic = \(asl,gl).
     (print_goal (asl,gl);
      ALL_TAC (asl,gl));;

let POP_TAC = POP_ASSUM (\th. ALL_TAC);;

let ETA_TAC v :tactic = \(asl,gl).
      let t1,t2 = dest_eq gl in
      let at1 = "\^v.^t1 ^v" in
      let at2 = "\^v.^t2 ^v" in
      [asl,mk_eq(at1,at2)],
      \[thm]. TRANS (SYM(ETA_CONV at1)) (TRANS thm (ETA_CONV at2));;

let EXT_TAC v :tactic = \(asl,gl).
      let t1,t2 = dest_eq gl in
      let at1 = mk_comb(t1,v) in
      let at2 = mk_comb(t2,v) in
      [asl,mk_forall(v,mk_eq(at1,at2))],
      \[thm]. EXT thm;;

let CHECK_TAC :tactic = \(asl,gl).
      if exists (aconv gl) asl
      then ACCEPT_TAC (ASSUME gl) (asl,gl)
      else failwith `CHECK_TAC`;;

let FALSE_TAC :tactic = \(asl,gl).
      if exists (\th. th = "F") asl
      then CONTR_TAC (ASSUME "F") (asl,gl)
      else failwith `CHECK_TAC`;;

let MP_IMP_TAC imp_th :tactic = \(asl,gl).
      let ant,cns = dest_imp(concl imp_th) in
      [asl,ant],
      \[thm]. MP imp_th thm;;

let UNASSUME_THEN ttac tm :tactic (asl,t) =
 if mem tm asl 
 then ttac (ASSUME tm) (subtract asl [tm], t)
 else failwith `UNASSUME_TAC`;;

let CONTRAPOS_TAC :tactic = \(asl,gl).
      let ant,cns = dest_imp gl in
      [asl,mk_imp (dest_neg cns, dest_neg ant)],
      \[thm]. CONTRAPOS thm;;

let FORALL_EQ_TAC :tactic = \(asl,gl).
     (let allt1,allt2 = dest_eq gl in
      let x,t1 = dest_forall allt1 in
      let y,t2 = dest_forall allt2 in
      if not (x = y) then fail
      else 
       [asl,mk_eq (t1, t2)],
       \[thm]. FORALL_EQ x thm)
      ? failwith `FORALL_EQ_TAC`;;

let EXISTS_EQ_TAC :tactic = \(asl,gl).
     (let ext1,ext2 = dest_eq gl in
      let x,t1 = dest_exists ext1 in
      let y,t2 = dest_exists ext2 in
      if not (x = y) then fail
      else 
       [asl,mk_eq (t1, t2)],
       \[thm]. EXISTS_EQ x thm)
      ? failwith `EXISTS_EQ_TAC`;;


% introducing a beta abstration

              A |- t
     ========================
           A |- (\x.t)x
%

% [PVH 95.2.19] %

let UNBETA_TAC x :tactic = \(asl,gl).
     ([asl,mk_comb(mk_abs(x, gl), x)],
       \[thm]. BETA_RULE thm)
      ? failwith `UNBETA_TAC`;;

%
g "x < y";;
e(UNBETA_TAC "x:num");;
e(UNBETA_TAC "y:num");;
%
%<
let WELL_FOUNDED_NUM_TAC :tactic = \(asl,gl).
     (let n = (fst o dest_forall) gl in
        (GEN_TAC
         THEN UNBETA_TAC n
         THEN SPEC_TAC(n,n)
         THEN MATCH_MP_TAC NUM_WELL_FOUNDED
         THEN BETA_TAC
         THEN GEN_TAC
         THEN DISCH_TAC)) (asl,gl)
      ? failwith `WELL_FOUNDED_NUM_TAC`;;
>%

%<
% ! implication abstraction

          A |- t1 ==> t2
     -----------------------
      A |- (!x.t1) ==> (!x.t2)
%

% Adapted from FORALL_EQ: [PVH 94.12.05]                                             %

let FORALL_IMP =
    \x. let all t = mk_forall(x,t) in
        \th. let (t1,t2) = dest_imp (concl th) in     %  t1 ==> t2  %
             let at1 = all t1 in
             DISCH at1 (GEN x
               (MATCH_MP (GEN x th) (SPEC x (ASSUME at1))))
        ? failwith `FORALL_IMP`;;

%
FORALL_IMP = - : (term -> thm -> thm)
FORALL_IMP "n:num" (SPEC_ALL SUB_ADD);;
FORALL_IMP "n:num" (SPEC_ALL (ASSUME "!n:num. R n ==> Q n"));;
%

let FORALL_IMP_TAC :tactic = \(asl,gl).
     (let allt1,allt2 = dest_imp gl in
      let x,t1 = dest_forall allt1 in
      let y,t2 = dest_forall allt2 in
      if not (x = y) then fail
      else 
       [asl,mk_imp (t1, t2)],
       \[thm]. FORALL_IMP x thm)
      ? failwith `FORALL_IMP_TAC`;;
>%

letrec UNDISCH_ALL_TAC :tactic = \(asl,gl).
       if asl = [] then ALL_TAC (asl,gl)
       else (UNDISCH_TAC (hd asl)
             THEN UNDISCH_ALL_TAC) (asl,gl);;

let UNDISCH_LAST_TAC :tactic =
    POP_ASSUM MP_TAC;;

let DEFINE_NEW_VAR def :tactic =
      let (v,tm) = dest_eq def in
      SUBGOAL_THEN (mk_exists(v,def)) STRIP_ASSUME_TAC
      THENL
        [  EXISTS_TAC tm
           THEN REWRITE_TAC[];
           ALL_TAC
        ]
      ? failwith `DEFINE_NEW_VAR`;;


let DOUBLE_LIST_INDUCT_TAC :tactic =
    LIST_INDUCT_TAC
    THENL
      [ ALL_TAC;
        GEN_TAC THEN LIST_INDUCT_TAC THEN POP_TAC
      ];;

let LENGTH_LIST_INDUCT_TAC :tactic =
    LIST_INDUCT_TAC
    THEN REWRITE_TAC[LENGTH;LENGTH_NIL;LENGTH_CONS]
    THEN REPEAT GEN_TAC THEN STRIP_TAC
    THEN POP_ASSUM (\th. REWRITE_TAC[th]);;


let FORALL_SYM_CONV :conv = \tm1.
      let x,body1 = dest_forall tm1 in
      let y,body  = dest_forall body1 in
      let tm2 = mk_forall(y,mk_forall(x,body)) in
      IMP_ANTISYM_RULE ((DISCH_ALL o GENL[y;x] o SPECL[x;y] o ASSUME) tm1)
                       ((DISCH_ALL o GENL[x;y] o SPECL[y;x] o ASSUME) tm2)
      ? failwith `FORALL_SYM_CONV`;;

%
FORALL_SYM_CONV "!x y. x+y=z";;
%


let NOT_AP_TERM_TAC f :tactic = \(asl,gl).
     (let eq_gl = dest_neg gl in
      let a,b = dest_eq eq_gl in
      [asl,mk_neg(mk_eq(mk_comb(f,a),mk_comb(f,b)))],
      \[thm]. MP (CONTRAPOS (DISCH eq_gl (AP_TERM f (ASSUME eq_gl)))) thm)
      ? failwith `NOT_AP_TERM_TAC`;;

let APP_let_CONV :conv = \tm1.
      let f,args = strip_comb tm1 in
      let lt.rest = args in
      let bodyf,val = dest_let lt in
      let var,body = dest_abs bodyf in
      let tm2 = mk_let((mk_abs(var,(list_mk_comb(f,(body.rest))))),val) in
      TRANS ((ONCE_DEPTH_CONV let_CONV) tm1)
            (SYM ((ONCE_DEPTH_CONV let_CONV) tm2))
      ? failwith `APP_LET_CONV`;;

let LIFT_let_TAC = (REPEAT o CONV_TAC o CHANGED_CONV o ONCE_DEPTH_CONV)
                    APP_let_CONV;;

let STRIP_let_TAC :tactic = \(asl,gl).
     (let bodyf,val = dest_let gl in
      let var,body = dest_abs bodyf in
      let let_equal = mk_eq(var,val) in
      [let_equal.asl,body],
      \[thm]. (pthm (SYM (let_CONV gl)); pthm thm; map (pthm o ASSUME) (hyp thm);
               (pthm o pthm) (EQ_MP (SYM (let_CONV gl))
                      (DISCH let_equal thm))))
      ? failwith `STRIP_let_TAC`;;

let USE_IMP_EQ_matches th =
     (let vars,imps = strip_forall(concl th) in
      let hyps,body = strip_imp imps in
      let left = fst(dest_eq body) in
      \tm. match left tm )
      ? failwith `USE_IMP_EQ_matches`;;

let USE_IMP_matches th =
     (let vars,imps = strip_forall(concl th) in
      let hyps,body = strip_imp imps in
      \tm. match body tm )
      ? failwith `USE_IMP_matches`;;

let SUB_matches (f:term->*) tm =
    if is_comb tm then
       (let rator,rand = dest_comb tm in
        (f rator ? f rand))
    if is_abs tm then
       (let bv,body = dest_abs tm in
        f body)
    else (failwith `SUB_matches`);;

letrec ONCE_DEPTH_matches (f:term->*) tm =
       (f tm ? ((SUB_matches (ONCE_DEPTH_matches f)) tm));;

let ONCE_DEPTH_USE_IMP_EQ_matches th =
    ONCE_DEPTH_matches (USE_IMP_EQ_matches th);;

let ONCE_DEPTH_USE_IMP_matches th =
    ONCE_DEPTH_matches (USE_IMP_matches th);;

let USE_IMP_THEN ttac th :tactic = \(asl,gl).
     (let matches = ONCE_DEPTH_USE_IMP_EQ_matches th gl in
      let ith = INST_TY_TERM matches (SPEC_ALL th) in
      let subgoal = concl (UNDISCH_ALL ith) in
       (SUBGOAL_THEN subgoal ttac
        THEN (MATCH_MP_TAC th ORELSE ALL_TAC)) (asl,gl) )
      ? failwith `USE_IMP_THEN`;;

let USE_IMP_TAC th :tactic = USE_IMP_THEN ASSUME_TAC th;;

let USE_IMP_AND_THEN ttac th tac :tactic =
    USE_IMP_THEN ttac th
    THENL [ tac; ALL_TAC ];;

let USE_THEN ttac th :tactic = \(asl,gl).
     (let matches = ONCE_DEPTH_USE_IMP_matches th gl in
      let ith = INST_TY_TERM matches (SPEC_ALL th) in
      let subgoal = concl (UNDISCH_ALL ith) in
       (SUBGOAL_THEN subgoal ttac
        THEN (MATCH_MP_TAC th ORELSE ALL_TAC)) (asl,gl) )
      ? failwith `USE_IMP_THEN`;;

let USE_TAC th :tactic = USE_THEN ASSUME_TAC th;;

let USE_AND_THEN ttac th tac :tactic =
    USE_THEN ttac th
    THENL [ tac; ALL_TAC ];;

let RES2_THEN thm_tac =
    RES_THEN (\th. IMP_RES_THEN thm_tac th ORELSE ALL_TAC);;

let IMP_RES2_THEN thm_tac =
    IMP_RES_THEN (\th. IMP_RES_THEN thm_tac th ORELSE ALL_TAC);;

let IMP_RES_M_THEN tac th g =
    IMP_RES_THEN (REPEAT_GTCL IMP_RES_THEN tac) th g ? ALL_TAC g;;

let RES_M_THEN tac g =
    RES_THEN (REPEAT_GTCL IMP_RES_THEN tac) g ? ALL_TAC g;;

let REWRITE_THM = \th. REWRITE_TAC[th];;
let ASM_REWRITE_THM = \th. ASM_REWRITE_TAC[th];;
let ONCE_REWRITE_THM = \th. ONCE_REWRITE_TAC[th];;
let IMP_RES_REWRITE_TAC = IMP_RES_THEN REWRITE_THM;;
let RES_REWRITE_TAC = RES_THEN REWRITE_THM;;

let REWRITE_ASSUM_TAC rths :tactic =
    RULE_ASSUM_TAC (REWRITE_RULE rths);;

let REWRITE_ALL_TAC rths :tactic =
    REWRITE_TAC rths  THEN
    REWRITE_ASSUM_TAC rths;;

let ASM_REWRITE_ALL_TAC rths :tactic =
    REWRITE_ASSUM_TAC rths  THEN
    ASM_REWRITE_TAC rths;;

let ONCE_REWRITE_ASSUM_TAC rths :tactic =
    RULE_ASSUM_TAC (ONCE_REWRITE_RULE rths);;

let ONCE_REWRITE_ALL_TAC rths :tactic =
    ONCE_REWRITE_TAC rths  THEN
    ONCE_REWRITE_ASSUM_TAC rths;;

let ONCE_ASM_REWRITE_ALL_TAC rths :tactic =
    ONCE_REWRITE_ASSUM_TAC rths  THEN
    ONCE_ASM_REWRITE_TAC rths;;

let REWRITE_ALL_THM = \th. REWRITE_ALL_TAC[th];;
let ASM_REWRITE_ALL_THM = \th. ASM_REWRITE_ALL_TAC[th];;
let ONCE_REWRITE_ALL_THM = \th. ONCE_REWRITE_ALL_TAC[th];;
let ONCE_ASM_REWRITE_ALL_THM = \th. ONCE_ASM_REWRITE_ALL_TAC[th];;


let UNIFY_VARS_TAC :tactic =
    ASSUM_LIST (\thms. MAP_EVERY (\th. REWRITE_ALL_TAC[th])
                 (mapfilter (SYM o (test ((pairf dest_var dest_var)
                                          o dest_eq o concl))) thms));;

