%
  Functions to help the VCG PL parser.
%

let mk_string s = mk_const(`\`` ^ s ^ `\``,":string");;

let mk_num n          = "NUM ^(mk_const(n,":num"))"
and mk_simple_var nm  = "VAR ^(mk_string(nm)) 0"
and mk_proc_name nm   = mk_string(nm)
and mk_pvar var       = "PVAR ^var"
and mk_inc var        = "INC  ^var"
and mk_plus (e1,e2)   = "PLUS ^e1 ^e2"
and mk_minus (e1,e2)  = "MINUS ^e1 ^e2"
and mk_mult (e1,e2)   = "MULT ^e1 ^e2"
and mk_null_exps ()   = "NIL:(exp)list"
and mk_exps (e,es)    = "CONS ^e ^es"
and mk_equ (e1,e2)    = "EQ ^e1 ^e2"
and mk_less (e1,e2)   = "LESS ^e1 ^e2"
and mk_lless (es1,es2) = "LLESS ^es1 ^es2"
and mk_not (b)        = "NOT ^b"
and mk_conj (b1,b2)   = "AND ^b1 ^b2"
and mk_disj (b1,b2)   = "OR ^b1 ^b2"
and mk_seq (c1,c2)    = "SEQ ^c1 ^c2"
and mk_skip ()        = "SKIP"
and mk_abort ()       = "ABORT"
and mk_assign (x,e)   = "ASSIGN ^x ^e"
and mk_if (b,c1,c2)   = "IF ^b ^c1 ^c2"
and mk_while (a,pr,b,c) = "WHILE ^a ^pr ^b ^c"
and mk_call (p,xs,es) = "CALL ^p ^xs ^es"
;;


let mk_anum n         = "ANUM ^(mk_const(n,":num"))"
and mk_avar var       = "AVAR ^var"
and mk_aplus (e1,e2)  = "APLUS ^e1 ^e2"
and mk_aminus (e1,e2) = "AMINUS ^e1 ^e2"
and mk_amult (e1,e2)  = "AMULT ^e1 ^e2"
and mk_null_vexps ()  = "NIL:(vexp)list"
and mk_vexps (e,es)   = "CONS ^e ^es"
and mk_aequ (e1,e2)   = "AEQ ^e1 ^e2"
and mk_aless (e1,e2)  = "ALESS ^e1 ^e2"
and mk_alless (es1,es2) = "ALLESS ^es1 ^es2"
and mk_aeqb (b1,b2)   = "AEQB ^b1 ^b2"
and mk_anot (b)       = "ANOT ^b"
and mk_true ()        = "ATRUE"
and mk_false ()       = "AFALSE"
and mk_aand (b1,b2)   = "AAND ^b1 ^b2"
and mk_aor (b1,b2)    = "AOR ^b1 ^b2"
and mk_aimp (b1,b2)   = "AIMP ^b1 ^b2"
and mk_acond (b1,b2,b3) = "ACOND ^b1 ^b2 ^b3"
and mk_close (b)      = "ACLOSE ^b"
and mk_aforall (x,a)  = "AFORALL ^x ^a"
and mk_aexists (x,a)  = "AEXISTS ^x ^a"
;;


let mk_empty_decl ()  = "DEMPTY"
and mk_proc (p,vars,vals,glbs,pre,post,calls,rec,c) =
                        "PROC ^p ^vars ^vals ^glbs ^pre ^post ^calls ^rec ^c"
and mk_decl_seq (d1,d2) = "DSEQ ^d1 ^d2"
and mk_null_vars ()   = "[]:(var)list"
and mk_vars (v,vs)    = "CONS ^v ^vs"
and mk_calls (p,a,cs) = "BND ^a ^p ^cs"
and mk_null_calls ()  = "\p:string. AFALSE"
and mk_null_rec ()    = "AFALSE"
and mk_program (d,c)  = "PROG ^d ^c"
;;

let mk_PCSPEC(p,c,q,d) =
                 "(r = mkenv ^d r0) ==> PCSPEC(^p,^c,^q,r)"
and mk_TCSPEC(p,c,d) =
                 "(r = mkenv ^d r0) ==> TCSPEC(^p,^c,r)"
and mk_CSPEC(p,c,q,d) =
                 "(r = mkenv ^d r0) ==> WFenv r ==> CSPEC(^p,^c,^q,r)"
and mk_PSPEC (p,a)    = "PSPEC(^p,^a)"
and mk_CPROGRESS(a,c,p,q,d) =
                 "(r = mkenv ^d r0) ==> CPROGRESS(^a,^c,^p,^q,r)"
and mk_PRE_PROGRESS(a,c,d) =
                 "(r = mkenv ^d r0) ==> PRE_PROGRESS(^a,^c,r)"
and mk_BODY_PROGRESS(a,c,calls,d) =
                 "(r = mkenv ^d r0) ==> BODY_PROGRESS(^a,^c,^calls,r)"
and mk_null_proc_names ()  = "NIL:(string)list"
and mk_proc_names (p,ps)   = "CONS ^p ^ps"
and mk_PATH_PROGRESS(a,p1,ps,p2,q,d) =
                 "(r = mkenv ^d r0) ==> PATH_PROGRESS(^a,^p1,^ps,^p2,^q,r)"
and mk_REC_PROGRESS(a,p,q,d) =
                 "(r = mkenv ^d r0) ==> REC_PROGRESS(^a,^p,^q,r)"
and mk_C_calls_TERM(a,c,d) =
                 "(r = mkenv ^d r0) ==> C_calls_TERM(^a,^c,r)"
and mk_M_calls_TERM(p,d) =
                 "(r = mkenv ^d r0) ==> M_calls_TERM(^p,r)"
;;
