% ================== %
% THE SUNRISE SYSTEM %
%       \ | /        %
%     ---_*_---      %
% ================== %

% This file contains the tactic VCG_TAC, which reduces a goal which is  %
% of the form PSPEC(pi,q), to the verification conditions yielded by    %
% the vcg, which are simplified with regards to programming-language    %
% constructs. %

% Set up autoloading from ancestor theories of `vcg`? %

%
loadf `autoload`;;

autoload_all_defs_and_thms
     `string bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules semantic_stages
      stage_semantics termination vcg`;;
%

let vcg_THM       = theorem    `vcg` `vcg_THM`;;
let vcg           = definition `vcg` `vcg`;;
let vcgg_THM      = theorem    `vcg` `vcgg_THM`;;
let vcgg          = definition `vcg` `vcgg`;;
let graph_vcs     = definition `vcg` `graph_vcs`;;
let extend_graph_vcs  = definition `vcg` `extend_graph_vcs`;;
let fan_out_graph_vcs = theorem    `vcg` `fan_out_graph_vcs`;;
let proc_names    = definition `vcg` `proc_names`;;
let vcgd_THM      = theorem    `vcg` `vcgd_THM`;;
let vcgd          = definition `vcg` `vcgd`;;
let vcgc_THM      = theorem    `vcg` `vcgc_THM`;;
let vcgc          = definition `vcg` `vcgc`;;
let vcg1_THM      = theorem    `vcg` `vcg1_THM`;;
let vcg1          = definition `vcg` `vcg1`;;
let dest_ALESS    = definition `vcg` `dest_ALESS`;;
let APPEND_INFIX  = definition `vcg` `APPEND_INFIX`;;
let call_path_progress = definition `progress_rules` `call_path_progress`;;
let call_progress = definition `progress_rules` `call_progress`;;
let mk_foralls    = definition `hoare_rules` `mk_foralls`;;
let WFc_WHILE     = theorem    `cmd_semantics` `WFc_WHILE`;;
let EMPTY_PROGRESS_ENV = definition `well_formed` `EMPTY_PROGRESS_ENV`;;
let WFenv_EMPTY_ENV = theorem  `well_formed` `WFenv_EMPTY_ENV`;;
let mk_eqs_THM    = theorem    `well_formed` `mk_eqs_THM`;;
let induct_pre    = definition `well_formed` `induct_pre`;;
let WFenv_WFenv_syntax = theorem `well_formed` `WFenv_WFenv_syntax`;;
let WFproc_syntax = definition `well_formed` `WFproc_syntax`;;
let WFproc_pre    = definition `well_formed` `WFproc_syntax`;;
let WFproc_calls  = definition `well_formed` `WFproc_syntax`;;
let WFproc_rec    = definition `well_formed` `WFproc_syntax`;;
let WFproc_correct = definition `well_formed` `WFproc_syntax`;;
let WFproc        = definition `well_formed` `WFproc`;;
let WFenv         = definition `well_formed` `WFenv`;;
let WFp           = definition `well_formed` `WFp`;;
let WFd           = definition `well_formed` `WFd`;;
let WFcalls       = definition `well_formed` `WFcalls`;;
let WFc           = definition `well_formed` `WFc`;;
let WFb           = definition `well_formed` `WFb`;;
let WFes          = definition `well_formed` `WFes`;;
let WFe           = definition `well_formed` `WFe`;;
let NOT_WFxs      = definition `well_formed` `NOT_WFxs`;;
let WFxs          = definition `well_formed` `WFxs`;;
let WFx           = definition `well_formed` `WFx`;;
let WFs           = definition `well_formed` `WFs`;;
let PSPEC         = definition `progress` `PSPEC`;;
let CSPEC_THM     = theorem    `progress` `CSPEC_THM`;;
let CLOSE         = definition `progress` `CLOSE`;;
let assign_list   = definition `translations` `assign_list`;;
let assign        = definition `translations` `assign`;;
let ab_pre        = definition `translations` `ab_pre`;;
let AB1_DEF       = definition `translations` `AB1_DEF`;;
let AB_DEF        = definition `translations` `AB_DEF`;;
let AB_state_DEF  = definition `translations` `AB_state_DEF`;;
let VES1_DEF      = definition `translations` `VES1_DEF`;;
let VES_DEF       = definition `translations` `VES_DEF`;;
let VES_state_DEF = definition `translations` `VES_state_DEF`;;
let VE1_DEF       = definition `translations` `VE1_DEF`;;
let VE_DEF        = definition `translations` `VE_DEF`;;
let VE_state_DEF  = definition `translations` `VE_state_DEF`;;
let trans_cont_THM = theorem   `translations` `trans_cont_THM`;;
let trans_cont    = definition `translations` `trans_cont`;;
let GVc           = definition `free_variables` `GVc`;;
let FVc_DEF       = definition `free_variables` `FVc_DEF`;;
let FVb_DEF       = definition `free_variables` `FVb_DEF`;;
let FVes_DEF      = definition `free_variables` `FVes_DEF`;;
let FVe_DEF       = definition `free_variables` `FVe_DEF`;;
let FVxs_DEF      = definition `free_variables` `FVxs_DEF`;;
let mkenv         = definition `prog_semantics` `mkenv`;;
let INITIAL_STATE = definition `prog_semantics` `INITIAL_STATE`;;
let EMPTY_ENV     = definition `prog_semantics` `EMPTY_ENV`;;
let ENV_TUPLE     = theorem    `prog_substitutions` `ENV_TUPLE`;;
let var_list_IDENT = theorem   `substitutions` `var_list_IDENT`;;
let var_list_THM  = theorem    `substitutions` `var_list_THM`;;
let a_subst       = definition `substitutions` `a_subst`;;
let vs_subst      = definition `substitutions` `vs_subst`;;
let v_subst       = definition `substitutions` `v_subst`;;
let a_subst_IDENT = theorem    `substitutions` `a_subst_IDENT`;;
let IDENT_SS_var  = theorem    `substitutions` `IDENT_SS_var`;;
let IDENT_SS      = definition `substitutions` `IDENT_SS`;;
let FVa_DEF       = definition `assert_semantics` `FVa_DEF`;;
let FVvs_DEF      = definition `assert_semantics` `FVvs_DEF`;;
let FVv_DEF       = definition `assert_semantics` `FVv_DEF`;;
let A_DEF         = definition `assert_semantics` `A_DEF`;;
let VS_DEF        = definition `assert_semantics` `VS_DEF`;;
let V_DEF         = definition `assert_semantics` `V_DEF`;;
let LEXI          = theorem    `assert_semantics` `LEXI`;;
let aexp_distinct = theorem    `assert_syntax` `aexp_distinct`;;
let aexp_one_one  = theorem    `assert_syntax` `aexp_one_one`;;
let vexp_distinct = theorem    `assert_syntax` `vexp_distinct`;;
let vexp_one_one  = theorem    `assert_syntax` `vexp_one_one`;;
let variant_increment = theorem `variants` `variant_increment`;;
let variants_THM  = theorem    `variants` `variants_THM`;;
let variants      = definition `variants` `variants`;;
let mk_variant    = definition `variants` `mk_variant`;;
let variant_ident = theorem    `variants` `variant_ident`;;
let variant_THM   = theorem    `variants` `variant_THM`;;
let logicals_APPEND = theorem  `variants` `logicals_APPEND`;;
let logicals_THM  = theorem    `variants` `logicals_THM`;;
let logicals      = definition `variants` `logicals`;;
let logical       = definition `variants` `logical`;;
let mk_logical    = definition `variants` `mk_logical`;;
let LOG_CHAR      = definition `variants` `LOG_CHAR`;;
let DISJOINT_LIST = definition `bindings` `DISJOINT_LIST`;;
let SL_APPEND     = theorem    `bindings` `SL_APPEND`;;
let SET_LIST      = definition `bindings` `SET_LIST`;;
let BNDsL         = theorem    `bindings` `BNDsL`;;
let BND_NOT_MATCH = theorem    `bindings` `BND_NOT_MATCH`;;
let BND_MATCH     = theorem    `bindings` `BND_MATCH`;;
let BND           = theorem    `bindings` `BND`;;
let var_one_one   = theorem    `variables` `var_one_one`;;
let UNION_THM     = theorem    `more_finite_sets` `UNION_THM`;;
let UNION_SET     = theorem    `more_finite_sets` `UNION_SET`;;
let IMAGE         = theorem    `more_finite_sets` `IMAGE`;;
let DELETE        = theorem    `more_finite_sets` `DELETE`;;
let UNION         = theorem    `more_finite_sets` `UNION`;;
let SUBSET        = theorem    `more_finite_sets` `SUBSET`;;
let IN            = theorem    `more_finite_sets` `IN`;;
let IN_UNION      = definition `finite_sets` `IN_UNION`;;
let UNION_COMM    = theorem    `finite_sets` `UNION_COMM`;;
let EMPTY_SUBSET  = theorem    `finite_sets` `EMPTY_SUBSET`;;
let ABSORPTION    = theorem    `finite_sets` `ABSORPTION`;;
let APPEND_NIL    = theorem    `list` `APPEND_NIL`;;
let APPEND        = definition `list` `APPEND`;;
let ALL_EL        = definition `list` `ALL_EL`;;
let o_THM         = theorem    `combin` `o_THM`;;



map delete_cache (words
     `combin list
      string bindings variables assert_syntax assert_semantics
      prog_syntax prog_semantics finite_sets more_finite_sets 
      variants free_variables substitutions var_substitutions
      prog_substitutions translations progress well_formed
      cmd_semantics hoare_rules progress_rules semantic_stages
      stage_semantics termination vcg`);;





% global controls of the vcg tactics: %

new_flag(`print_vcg`,false) ? ();;

letref assignable_print_vcg_term = print_term;;


% New conversions for rewriting: %

let debug_CONV tm =
   print_term tm; print_newline(); print_newline();
   ALL_CONV tm;;


let check_op_CONV name nargs =
  \tm. let (op,args) = strip_comb tm in
       if not (fst(dest_const op) = name) then fail
       else if not (length args = nargs) then fail
       else ALL_CONV tm ;;

%
let (op,args) = (strip_comb o snd o top_goal) ();;
(fst o dest_const)op;;
%

let var_EQ_CONV =
  REWR_CONV var_one_one
  THENC RATOR_CONV (RAND_CONV string_EQ_CONV)
  THENC RAND_CONV num_EQ_CONV
  THENC PURE_ONCE_REWRITE_CONV[AND_CLAUSES];;

% Tests of var_EQ_CONV:
var_EQ_CONV "(VAR `x` 2) = (VAR `x` 2)";;
var_EQ_CONV "(VAR `x` 2) = (VAR `y` 2)";;
var_EQ_CONV "(VAR `x` 2) = (VAR `x` 0)";;
var_EQ_CONV "(VAR `x` 2) = (VAR `y` 0)";;
var_EQ_CONV "(VAR `x` 2000) = (VAR `y` 0)";;
var_EQ_CONV "(VAR `xx` 2) = (VAR `y` 0)";;
var_EQ_CONV "(VAR `abcdefghijklmnopqrstuvwxyz` 2) = (VAR `y` 0)";;
(DEPTH_CONV var_EQ_CONV)
             "(VAR `x` 2 = VAR `x` 2) \/ (VAR `x` 2 = VAR `x` 0) /\
              (VAR `x` 2 = VAR `x` 2) \/ (VAR `x` 2 = VAR `x` 0) /\
              (VAR `x` 2 = VAR `x` 2) \/ (VAR `x` 2 = VAR `x` 0) /\
              (VAR `x` 2 = VAR `x` 2) \/ (VAR `x` 2 = VAR `x` 0) ";;
%

let BND_CONV eq_conv =
  \tm. let (bnd,[val;key;base;obj]) = strip_comb tm in
       (if not((fst o dest_const)bnd = `BND`) then fail;
        if obj = key then ISPECL[val;key;base] BND_MATCH
        else let notkey = EQF_ELIM(eq_conv(mk_eq(obj,key))) in
             (MP (ISPECL[val;key;base;obj] BND_NOT_MATCH) notkey));;


let IDENT_SS_CONV = REWR_CONV IDENT_SS_var;;

%
IDENT_SS_CONV "IDENT_SS (VAR `x` 0)";;
IDENT_SS_CONV "(VAR `x` 0)";;
%

let var_BND_CONV = DEPTH_CONV (BND_CONV var_EQ_CONV ORELSEC IDENT_SS_CONV);;

% Tests of var_BND_CONV:
var_BND_CONV "BND (ANUM 3) (VAR `x` 2) IDENT_SS (VAR `x` 2)";;
var_BND_CONV "(BND (ANUM 3) (VAR `x` 2)
               (BND (ANUM 5) (VAR `y` 0) IDENT_SS)) (VAR `y` 0)";;
var_BND_CONV
             "(BND (ANUM 3) (VAR `x` 0)
               (BND (ANUM 4) (VAR `y` 0)
                (BND (ANUM 5) (VAR `z` 0) IDENT_SS))) (VAR `z` 0)";;
var_BND_CONV
             "(BND (ANUM 3) (VAR `x` 0)
               (BND (ANUM 4) (VAR `y` 0)
                (BND (ANUM 5) (VAR `z` 0) IDENT_SS))) (VAR `w` 0)";;
%

let string_BND_CONV = DEPTH_CONV (BND_CONV string_EQ_CONV);;

let WFx_CONV =
         REWR_CONV WFx THENC
         RAND_CONV string_CONV THENC
         PURE_ONCE_REWRITE_CONV[WFs] THENC
         RAND_CONV (RAND_CONV (PURE_ONCE_REWRITE_CONV[LOG_CHAR])) THENC
         RAND_CONV ascii_EQ_CONV THENC
         PURE_ONCE_REWRITE_CONV[NOT_CLAUSES];;

%
WFx_CONV "WFx (VAR `abc` 3)";;
%

let check_const name const =
    if (name = (fst (dest_const const))) then true
    else failwith (`not `^name) ;;

% ----------------------------------------------------------------------- %
% WFxs_CONV : conv -> conv                                                %
% WFxs_CONV conv "WFxs [x0;...;xn]" --->                                  %
% |- WFxs [x0;...;xn] = T iff WFx_CONV "WFx xi"---> |- P xi = T for all i %
% |- WFxs [x0;...;xn] = F otherwise                                       %
% ----------------------------------------------------------------------- %
let WFxs_CONV =
    let bth,ith = CONJ_PAIR WFxs in
    let AND_THM = setify(flat(map (CONJ_LIST 5)
        [(SPEC "T" AND_CLAUSES);(SPEC "F" AND_CLAUSES)])) in
  \tm.
    (let (_,l) = ((check_const`WFxs`) # I)(dest_comb tm) in
     let lis = fst(dest_list l) in
     let ffn x th =
        let (left,right) = dest_eq(concl th) in
        let (_,ls) = dest_comb left in
        let fthm = SPECL [x;ls] ith and cthm = WFx_CONV "WFx ^x" in
        SUBS AND_THM (SUBS[cthm;th]fthm)
     in
     (itlist ffn lis bth)) ?\s failwith (`WFxs_CONV: `^s);;

%
WFxs_CONV "WFxs[VAR `n` 0;VAR `a` 0;VAR `b` 0]";;
WFxs_CONV "WFxs[VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0]";;
WFxs_CONV "WFxs[VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `n` 0;VAR `a` 0;VAR `b` 0;
                VAR `n` 0;VAR `a` 0;VAR `b` 0;VAR `^n` 0;VAR `a` 0;VAR `b` 0]";;
%


letrec REDUCE_CONV conv tm =
  (SUB_CONV (REDUCE_CONV conv)
   THENC (TRY_CONV (INSERT_CONV conv))) tm;;

%
REDUCE_CONV num_EQ_CONV ("1 INSERT (2 INSERT {})");;
REDUCE_CONV num_EQ_CONV ("{1,2,1}");;
REDUCE_CONV num_EQ_CONV ("{1,2,1,2,1}");;
%

let variant_CONV =
    REWR_CONV variant_THM THENC
    RATOR_CONV (RATOR_CONV (RAND_CONV (IN_CONV var_EQ_CONV))) THENC
    REWRITE_CONV[mk_variant] THENC
    ONCE_DEPTH_CONV ADD_CONV;;

%
variant_CONV ("variant (VAR `x` 0) { VAR `x` 0, VAR `x` 1 }");;
variant_CONV ("variant (VAR `x` 1) { VAR `x` 0, VAR `x` 1 }");;
variant_CONV ("variant (VAR `x` 2) { VAR `x` 0, VAR `x` 1 }");;
(DEPTH_CONV variant_CONV) ("variant (VAR `x` 0) { VAR `x` 0, VAR `x` 1 }");;
%


letrec variants_CONV =
  \tm. let (vrnts,[xs;s]) = strip_comb tm in
       (if not (vrnts = "variants") then fail;
        if xs = "[]:(var)list" then
          SPEC s (CONJUNCT1 variants)
        else
          (REWR_CONV (CONJUNCT2 variants) THENC
           RAND_CONV (DEPTH_CONV variant_CONV) THENC
           let_CONV THENC
           RAND_CONV variants_CONV) tm
        );;

%
variants_CONV ("variants [VAR `x` 0;VAR `y` 0;VAR `x` 1]
                       {VAR `x` 0,VAR `x` 1}");;
variants_CONV ("variants [VAR `x` 5;VAR `y` 0;VAR `x` 1]
                       {VAR `x` 0,VAR `x` 1}");;
variants_CONV ("variants []
                       {VAR `x` 0,VAR `x` 1}");;
%

let mk_logical_CONV =
  \tm. let (mlog,[s]) = strip_comb tm in
       (if not (mlog = "mk_logical") then fail;
        let (quote.rest) = (explode o fst o dest_const)s in
        let s1 = mk_const(implode(quote.`^`.rest),":string") in
        let th = SPEC s mk_logical in
            TRANS (REWRITE_RULE[LOG_CHAR] th) (SYM(string_CONV s1))
        );;

%
mk_logical_CONV ("mk_logical `x`");;
mk_logical_CONV ("mk_logical `abcdefghijklmopqrstuvwxyz`");;
%

let logical_CONV =
    REWR_CONV logical THENC
    RATOR_CONV (RAND_CONV mk_logical_CONV);;

%
logical_CONV ("logical (VAR `x` 0)");;
logical_CONV ("logical (VAR `abcdefghijklmopqrstuvwxyz` 12)");;
logical_CONV ("[]:(num)list");;
(DEPTH_CONV logical_CONV) ("[logical (VAR `x` 0);logical (VAR `x` 0);
                             logical (VAR `x` 0);logical (VAR `x` 0)]");;
%

%  logicals  |- logicals = MAP logical  %

let logicals_CONV =
    REWR_CONV (AP_THM logicals "vs:(var)list") THENC
    MAP_CONV logical_CONV;;

%
logicals_CONV ("logicals [VAR `x` 0;VAR `y` 0;VAR `x` 1]");;
logicals_CONV ("logicals [VAR `x` 0;VAR `y` 0;VAR `x` 1;VAR `x` 1;VAR `z` 0;
                          VAR `x` 0;VAR `y` 0;VAR `x` 1;VAR `x` 1;VAR `z` 0;
                          VAR `x` 0;VAR `y` 0;VAR `x` 1;VAR `x` 1;VAR `z` 0;
                          VAR `x` 0;VAR `y` 0;VAR `x` 1;VAR `x` 1;VAR `z` 0;
                          VAR `x` 0;VAR `y` 0;VAR `x` 1;VAR `x` 1;VAR `z` 0]");;
%

let GVc_CONV =
         check_op_CONV `GVc` 2 THENC
         PURE_REWRITE_CONV[GVc] THENC
         PURE_REWRITE_CONV[o_THM;mkenv] THENC
         string_BND_CONV THENC
         PURE_REWRITE_CONV[FST;SND] THENC
         %DEPTH_CONV APPEND_CONV THENC%
         PURE_REWRITE_CONV[SET_LIST] THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV);;

%
APPEND_CONV "APPEND [] [3]";;
GVc_CONV "GVc ^([[ if x<5 then b:=a+b-car else skip fi ]]) r";;


let c = [[ f(;x);g(;y) ]]
and r = [[
   procedure f(;val n);
      global a,b;
      pre   true;
      post  a = ^n * 2;

      a := n + n
   end procedure;

   procedure g(;val n);
      global a,b,c;
      pre   true;
      post  a = ^n * 3;

      f(;n); a := a + n
   end procedure
      ]] in
GVc_CONV "GVc ^c (mkenv ^r r0)";;
%

let FVc_CONV =
         check_op_CONV `FVc` 2 THENC
         PURE_REWRITE_CONV[FVc_DEF;o_THM;mkenv] THENC
         string_BND_CONV THENC
         REWRITE_CONV[] THENC
         ONCE_DEPTH_CONV let_CONV THENC
         PURE_REWRITE_CONV[FVes_DEF;FVb_DEF;FVe_DEF;FVxs_DEF] THENC
         PURE_REWRITE_CONV[SET_LIST] THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV) THENC
         REDUCE_CONV var_EQ_CONV;;

%
FVc_CONV "FVc ^([[ if x<5 then b:=a+b-car else skip fi ]]) r";;

let c = [[ f(;x);g(;y) ]]
and r = [[
   procedure f(;val n);
      global a,b;
      pre   true;
      post  a = ^n * 2;

      a := n + n
   end procedure;

   procedure g(;val n);
      global a,b,c;
      pre   true;
      post  a = ^n * 3;

      f(;n); a := a + n
   end procedure
      ]] in
FVc_CONV "FVc ^c (mkenv ^r r0)";;
%

let FVv_CONV =
         check_op_CONV `FVv` 1 THENC
         PURE_REWRITE_CONV[FVv_DEF] THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV ORELSEC
                     DELETE_CONV var_EQ_CONV);;

let FVa_CONV =
         check_op_CONV `FVa` 1 THENC
         PURE_REWRITE_CONV[FVa_DEF;FVv_DEF] THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV ORELSEC
                     DELETE_CONV var_EQ_CONV);;

%
FVa_CONV
    "FVa (AAND (AEXISTS (VAR `b` 0)
                   (AEQ (AVAR (VAR `^n` 0))
                        (APLUS (AMULT (ANUM 2) (AVAR (VAR `b` 0)))
                               (AVAR (VAR `a` 0)))))
         (AAND (ALESS (AVAR (VAR `a` 0)) (ANUM 2))
               (AEQ (AVAR (VAR `n` 0)) (AVAR (VAR `^n` 0)))
         ))";;
FVa_CONV
    "FVa ^([[ {(?b. ^n = 2*b + a) /\ a < 2 /\ n = ^n} ]])";;
g "zz = FVa ^([[ {(?b. ^n = 2*b + a) /\ a < 2 /\ n = ^n} ]])";;
e(CONV_TAC (RAND_CONV FVa_CONV));;
%



let NOT_IN_FVa_BND_PROGRESS_ENV =
 TAC_PROOF
  (([],
    "!p x a (p1:string) g.
       (~(x IN FVa(BND a p1 g p))) =
       ( (p = p1) ==> ~(x IN FVa a)) /\
       (~(p = p1) ==> ~(x IN FVa(g p)))"),
   REPEAT GEN_TAC
   THEN REWRITE_TAC[BND]
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[]
  );;

% NOT_IN_FVa_BND_PROGRESS_ENV
g "!p x a (p1:string) g.
       (~(x IN FVa(BND a p1 g p))) =
       ( (p = p1) ==> ~(x IN FVa a)) /\
       (~(p = p1) ==> ~(x IN FVa(g p)))";;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
%

let EMPTY_PROGRESS_ENV_THM =
 TAC_PROOF
  (([],
    "!p. g0 p = ATRUE"),
   REWRITE_TAC[EMPTY_PROGRESS_ENV]
  );;

% EMPTY_PROGRESS_ENV_THM
g "!p. g0 p = ATRUE";;
e(REWRITE_TAC[EMPTY_PROGRESS_ENV]);;
%

let NULL_PROGRESS_ENV =
 TAC_PROOF
  (([],
    "!p:string. (\p. AFALSE) p = AFALSE"),
   BETA_TAC
   THEN REWRITE_TAC[]
  );;

% NULL_PROGRESS_ENV
g "!p:string. (\p. AFALSE) p = AFALSE";;
e(BETA_TAC);;
e(REWRITE_TAC[]);;
%

let SOME_STRING_EQ =
 TAC_PROOF
  (([],
    "!p1:string.
       (!p. ~(p = p1)) = F"),
   CONV_TAC (ONCE_DEPTH_CONV FORALL_NOT_CONV)
   THEN REWRITE_TAC[]
   THEN GEN_TAC
   THEN EXISTS_TAC "p1:string"
   THEN REWRITE_TAC[]
  );;

% SOME_STRING_EQ
g "!p1:string.
       (!p. ~(p = p1)) = F";;
e(CONV_TAC (ONCE_DEPTH_CONV FORALL_NOT_CONV));;
e(REWRITE_TAC[]);;
e(GEN_TAC);;
e(EXISTS_TAC "p1:string");;
e(REWRITE_TAC[]);;
%



let [WFc_SKIP;WFc_ABORT;WFc_ASSIGN;WFc_SEQ;WFc_IF;WFc_WHILE1;WFc_CALL]
    = CONJUNCTS WFc;;


let WFc_CONV =
         check_op_CONV `WFc` 3 THENC
         PURE_REWRITE_CONV[WFc_SKIP;WFc_ABORT;WFc_ASSIGN;
                           WFc_SEQ;WFc_IF;WFc_WHILE;WFc_CALL] THENC
         PURE_REWRITE_CONV[mkenv] THENC
         PURE_REWRITE_CONV[o_THM] THENC
         string_BND_CONV THENC
         PURE_REWRITE_CONV[FST;SND] THENC
         PURE_REWRITE_CONV[NOT_IN_FVa_BND_PROGRESS_ENV] THENC
         PURE_REWRITE_CONV[EMPTY_PROGRESS_ENV_THM;NULL_PROGRESS_ENV] THENC
         DEPTH_CONV FVa_CONV THENC
         DEPTH_CONV FVv_CONV THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV) THENC
         DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
         REWRITE_CONV[SOME_STRING_EQ] THENC
         DEPTH_CONV LENGTH_CONV THENC
         REWRITE_CONV[] THENC
         PURE_REWRITE_CONV[WFb;WFes;WFe] THENC
         ONCE_DEPTH_CONV WFxs_CONV THENC
         DEPTH_CONV WFx_CONV THENC
         PURE_REWRITE_CONV[SET_LIST;DISJOINT_LIST;APPEND] THENC
         ONCE_DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
         REWRITE_CONV[];;

%
let WFc_CONV =
         check_op_CONV `WFc` 3 THENC
debug_CONV THENC
         PURE_REWRITE_CONV[WFc_SKIP;WFc_ABORT;WFc_ASSIGN;
                           WFc_SEQ;WFc_IF;WFc_WHILE;WFc_CALL] THENC
debug_CONV THENC
         PURE_REWRITE_CONV[mkenv] THENC
debug_CONV THENC
         PURE_REWRITE_CONV[o_THM] THENC
debug_CONV THENC
         string_BND_CONV THENC
debug_CONV THENC
         PURE_REWRITE_CONV[FST;SND] THENC
         PURE_REWRITE_CONV[NOT_IN_FVa_BND_PROGRESS_ENV] THENC
         PURE_REWRITE_CONV[EMPTY_PROGRESS_ENV_THM;NULL_PROGRESS_ENV] THENC
debug_CONV THENC
         DEPTH_CONV FVa_CONV THENC
         DEPTH_CONV FVv_CONV THENC
         DEPTH_CONV (UNION_CONV var_EQ_CONV) THENC
         DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
         REWRITE_CONV[SOME_STRING_EQ] THENC
debug_CONV THENC
         DEPTH_CONV LENGTH_CONV THENC
debug_CONV THENC
         REWRITE_CONV[] THENC
debug_CONV THENC
         PURE_REWRITE_CONV[WFb;WFes;WFe] THENC
debug_CONV THENC
         ONCE_DEPTH_CONV WFxs_CONV THENC
debug_CONV THENC
         DEPTH_CONV WFx_CONV THENC
debug_CONV THENC
         PURE_REWRITE_CONV[SET_LIST;DISJOINT_LIST;APPEND] THENC
debug_CONV THENC
         ONCE_DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
debug_CONV THENC
         REWRITE_CONV[];;

WFc_CONV "WFc (ASSIGN (VAR `x` 0) (PLUS (PVAR (VAR `y` 1)) (NUM 5))) g0 r";;
[[ x := y+5 ]];;

let c = [[ triangle(;5) ]]
and r = [[
   procedure triangle(;val n);
      global a;
      pre   true;
      post  2 * a = ^n * (^n + 1);

      if n = 0 then a := 0
      else
         triangle(;n - 1);
         a := a + n
      fi
   end procedure
      ]] in
WFc_CONV "WFc ^c g0 (mkenv ^r r0)";;

let c =
     [[ z := 0;
        assert x = x0 /\ y = z + y0
          with x - z < ^t
        while ~(z = x) do
           y := y+1; z := z+1
        od ]]
and r = [[
   procedure triangle(;val n);
      global a;
      pre   true;
      post  2 * a = ^n * (^n + 1);

      if n = 0 then a := 0
      else
         triangle(;n - 1);
         a := a + n
      fi
   end procedure
      ]] in
WFc_CONV "WFc ^c g0 (mkenv ^r r0)";;

let c =
     [[ z := 0;
        assert x = x0 /\ y = z + y0
          with x - z < ^t
        while ~(z = x) do
           y := y+1; z := z+1
        od ]]
and r = [[
   procedure triangle(;val n);
      global a;
      pre   true;
      post  2 * a = ^n * (^n + 1);

      if n = 0 then a := 0
      else
         triangle(;n - 1);
         a := a + n
      fi
   end procedure
      ]] in
WFc_CONV "WFc ^c (BND ^([[ {x+n < ^x} ]]) `p1`
                  (BND ^([[ {a-b < ^n} ]]) `triangle` (\p. AFALSE)))
                 (mkenv ^r r0)";;

let c =
     [[ z := 0;
        assert x = x0 /\ y = z + y0
          with x - z < ^t
        while ~(z = x) do
           y := y+1; z := z+1
        od ]]
and r = [[
   procedure triangle(;val n);
      global a;
      pre   true;
      post  2 * a = ^n * (^n + 1);

      if n = 0 then a := 0
      else
         triangle(;n - 1);
         a := a + n
      fi
   end procedure
      ]] in
WFc_CONV "WFc ^c (BND ^([[ {t < ^t} ]]) `p1` (\p. AFALSE)) (mkenv ^r r0)";;
%



let IMP_EQ_DIST =
 TAC_PROOF
  (([],
    "!a b c. (a ==> (b = c)) = ((a ==> b) = (a ==> c))"),
   REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
    [ DISCH_TAC
      THEN EQ_TAC
      THEN REPEAT STRIP_TAC
      THEN RES_TAC;

      DISCH_THEN (STRIP_ASSUME_TAC o uncurry CONJ o EQ_IMP_RULE)
      THEN DISCH_TAC
      THEN EQ_TAC
      THEN DISCH_TAC
      THEN UNDISCH_TAC "a:bool"
      THEN FIRST_ASSUM MATCH_MP_TAC
      THEN POP_ASSUM REWRITE_THM
    ]
  );;

%
g "!a b c. (a ==> (b = c)) = ((a ==> b) = (a ==> c))";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(DISCH_TAC);;
 e(EQ_TAC);;

  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;

  e(REPEAT STRIP_TAC);;
  e(RES_TAC);;

 e(DISCH_THEN (STRIP_ASSUME_TAC o uncurry CONJ o EQ_IMP_RULE));;
 e(DISCH_TAC);;
 e(EQ_TAC);;

  e(DISCH_TAC);;
  e(UNDISCH_TAC "a:bool");;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(POP_ASSUM REWRITE_THM);;

  e(DISCH_TAC);;
  e(UNDISCH_TAC "a:bool");;
  e(FIRST_ASSUM MATCH_MP_TAC);;
  e(POP_ASSUM REWRITE_THM);;
%


let BND_MATCH1 = (GENL["x:**";"e:*"] o SPECL["e:*";"x:**"]) BND_MATCH;;

let IMP_BND_MATCH_CONV tm =
    let (ant,con) = dest_imp tm in
    let (var,cnst) = dest_eq ant in
    let BND_th = ISPEC cnst BND_MATCH1 in
    PURE_ONCE_REWRITE_RULE[IMP_EQ_DIST]
      (DISCH_ALL (PURE_REWRITE_CONV[ASSUME ant;BND_th] con));;

%
IMP_BND_MATCH_CONV "(s = `f`) ==> (BND AFALSE `f` g s = ATRUE)";;
%


let IMP_BND_NOT_MATCH_CONV tm =
    let (ant,con) = dest_imp tm in
    let (var,cnst) = dest_eq (dest_neg ant) in
    let BND_th = MATCH_MP BND_NOT_MATCH (ASSUME ant) in
    PURE_ONCE_REWRITE_RULE[IMP_EQ_DIST]
      (DISCH_ALL (ADD_ASSUM ant (PURE_REWRITE_CONV[BND_th] con)));;

letrec ONCE_DEEP_IMP_BND_NOT_MATCH_CONV tm =
    (ONCE_DEPTH_CONV
      (IMP_BND_NOT_MATCH_CONV THENC
       RAND_CONV ONCE_DEEP_IMP_BND_NOT_MATCH_CONV)) tm;;

let DEEP_IMP_BND_NOT_MATCH_CONV =
    REPEATC (CHANGED_CONV ONCE_DEEP_IMP_BND_NOT_MATCH_CONV);;

%
IMP_BND_NOT_MATCH_CONV "~(s = `f`) ==> (BND AFALSE `f` g s = ATRUE)";;
DEEP_IMP_BND_NOT_MATCH_CONV
ONCE_DEEP_IMP_BND_NOT_MATCH_CONV
     "(~(s = `g`) ==>
       ((s = `f`) ==>
        let x' = APPEND(APPEND[][VAR `n` 0])[VAR `a` 0;VAR `b` 0]
        in
         (FVa(BND(AEQ(AVAR(VAR `n` 0))(AVAR(VAR `^n` 0)))`f` g0 s)) SUBSET
         (SL(APPEND x' x0))) /\
       (~(s = `f`) ==>
        (FVa(BND(AEQ(AVAR(VAR `n` 0))(AVAR(VAR `^n` 0)))`f`
             (BND(AEQ(AVAR(VAR `x` 0))(AVAR(VAR `^y` 0)))`g` (\p.AFALSE)) s))
        SUBSET
        (SL x0)))";;
%


let FVa_g0_SUBSET =
 TAC_PROOF
  (([],
    "!p t. FVa(g0 p) SUBSET t"),
   REWRITE_TAC[EMPTY_PROGRESS_ENV_THM;FVa_DEF;EMPTY_SUBSET]
  );;

%
g "!p t. FVa(g0 p) SUBSET t";;
e(REWRITE_TAC[EMPTY_PROGRESS_ENV_THM;FVa_DEF;EMPTY_SUBSET]);;
%

let FVa_NULL_ENV_SUBSET =
 TAC_PROOF
  (([],
    "!p t. FVa((\p:string.AFALSE) p) SUBSET t"),
   REWRITE_TAC[NULL_PROGRESS_ENV;FVa_DEF;EMPTY_SUBSET]
  );;

%
g "!p t. FVa((\p:string.AFALSE) p) SUBSET t";;
e(REWRITE_TAC[NULL_PROGRESS_ENV;FVa_DEF;EMPTY_SUBSET]);;
%

let BND_ENV_FVa_CALLS_SUBSET =
 TAC_PROOF
  (([],
    "!p p1 (r:env) g x0 vars vals glbs pre post calls rec c.
      (let (vars',vals',glbs',pre',post',calls',rec',c') =
           BND (vars,vals,glbs,pre,post,calls,rec,c) p1 r p
       in
       let x' = APPEND (APPEND vars' vals') glbs' in
       FVa(g p) SUBSET SL(APPEND x' x0)) =
      ( (p = p1) ==>
        (let x' = APPEND (APPEND vars vals) glbs in
         FVa(g p) SUBSET SL(APPEND x' x0))
      )
      /\
      (~(p = p1) ==>
        (let (vars',vals',glbs',pre',post',calls',rec',c') = r p in
         let x' = APPEND (APPEND vars' vals') glbs' in
         FVa(g p) SUBSET SL(APPEND x' x0))
      )"),
   REPEAT GEN_TAC
   THEN REWRITE_TAC[BND]
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[]
  );;

% BND_ENV_FVa_CALLS_SUBSET
g "!p p1 (r:env) g x0 vars vals glbs pre post calls rec c.
      (let (vars',vals',glbs',pre',post',calls',rec',c') =
           BND (vars,vals,glbs,pre,post,calls,rec,c) p1 r p
       in
       let x' = APPEND (APPEND vars' vals') glbs' in
       FVa(g p) SUBSET SL(APPEND x' x0)) =
      ( (p = p1) ==>
        (let x' = APPEND (APPEND vars vals) glbs in
         FVa(g p) SUBSET SL(APPEND x' x0))
      )
      /\
      (~(p = p1) ==>
        (let (vars',vals',glbs',pre',post',calls',rec',c') = r p in
         let x' = APPEND (APPEND vars' vals') glbs' in
         FVa(g p) SUBSET SL(APPEND x' x0))
      )";;
e(REPEAT GEN_TAC);;
e(REWRITE_TAC[BND]);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;
 e(CONV_TAC (DEPTH_CONV let_CONV));;
 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
%


let EMPTY_ENV_FVa_CALLS_SUBSET =
 TAC_PROOF
  (([],
    "!p g x0.
      (let (vars',vals',glbs',pre',post',calls',rec',c') = r0 p in
       let x' = APPEND (APPEND vars' vals') glbs' in
       FVa(g p) SUBSET SL(APPEND x' x0)) =
      (FVa(g p) SUBSET SL x0)"),
   REWRITE_TAC[EMPTY_ENV]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REWRITE_TAC[APPEND_NIL]
  );;

% EMPTY_ENV_FVa_CALLS_SUBSET
g "!p g x0.
      (let (vars',vals',glbs',pre',post',calls',rec',c') = r0 p in
       let x' = APPEND (APPEND vars' vals') glbs' in
       FVa(g p) SUBSET SL(APPEND x' x0)) =
      (FVa(g p) SUBSET SL x0)";;
e(REWRITE_TAC[EMPTY_ENV]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REWRITE_TAC[APPEND_NIL]);;
%


let SOME_ALESS_REC =
 TAC_PROOF
  (([],
    "!v' y' x.
       (?v y. (ALESS v' (AVAR y') = ALESS v (AVAR y)) /\
              ~WFx y /\ FVv v SUBSET SL x) =
       (~WFx y' /\ FVv v' SUBSET SL x)"),
   REPEAT GEN_TAC
   THEN EQ_TAC
   THENL
    [ REWRITE_TAC[aexp_one_one;vexp_one_one]
      THEN STRIP_TAC
      THEN ASM_REWRITE_TAC[];

      STRIP_TAC
      THEN EXISTS_TAC "v':vexp"
      THEN EXISTS_TAC "y':var"
      THEN ASM_REWRITE_TAC[]
    ]
  );;

% SOME_ALESS_REC
g "!v' y' x.
       (?v y. (ALESS v' (AVAR y') = ALESS v (AVAR y)) /\
              ~WFx y /\ FVv v SUBSET SL x) =
       (~WFx y' /\ FVv v' SUBSET SL x)";;
e(REPEAT GEN_TAC);;
e(EQ_TAC);;

 e(REWRITE_TAC[aexp_one_one;vexp_one_one]);;
 e(STRIP_TAC);;
 e(ASM_REWRITE_TAC[]);;

 e(STRIP_TAC);;
 e(EXISTS_TAC "v':vexp");;
 e(EXISTS_TAC "y':var");;
 e(ASM_REWRITE_TAC[]);;
%



let WFd_CONV =
         check_op_CONV `WFd` 2 THENC
         PURE_REWRITE_CONV[WFd;WFproc_syntax;mkenv] THENC
         ONCE_DEPTH_CONV WFc_CONV THENC
         ONCE_DEPTH_CONV GVc_CONV THENC
         ONCE_DEPTH_CONV FVc_CONV THENC
         PURE_REWRITE_CONV[BND_ENV_FVa_CALLS_SUBSET] THENC
         PURE_REWRITE_CONV[EMPTY_ENV_FVa_CALLS_SUBSET] THENC
         DEEP_IMP_BND_NOT_MATCH_CONV THENC
         ONCE_DEPTH_CONV IMP_BND_MATCH_CONV THENC
         string_BND_CONV THENC
         PURE_REWRITE_CONV[FVa_g0_SUBSET;FVa_NULL_ENV_SUBSET] THENC
         PURE_REWRITE_CONV[SOME_ALESS_REC] THENC
         REWRITE_CONV[aexp_distinct;vexp_distinct] THENC
         ONCE_DEPTH_CONV FVa_CONV THENC
         ONCE_DEPTH_CONV FVv_CONV THENC
         PURE_REWRITE_CONV[APPEND] THENC
         DEPTH_CONV let_CONV THENC
         ONCE_DEPTH_CONV logicals_CONV THENC
         PURE_REWRITE_CONV[APPEND] THENC
         PURE_REWRITE_CONV[DISJOINT_LIST;SUBSET;SET_LIST] THENC
         ONCE_DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
         REWRITE_CONV[] THENC
         ONCE_DEPTH_CONV WFxs_CONV THENC
         ONCE_DEPTH_CONV WFx_CONV THENC
         REWRITE_CONV[];;

%<

let WFd_CONV =
         check_op_CONV `WFd` 2 THENC
       debug_CONV THENC
         PURE_REWRITE_CONV[WFd;WFproc_syntax;mkenv] THENC
       debug_CONV THENC
         ONCE_DEPTH_CONV WFc_CONV THENC
         ONCE_DEPTH_CONV GVc_CONV THENC
         ONCE_DEPTH_CONV FVc_CONV THENC
       debug_CONV THENC
         PURE_REWRITE_CONV[BND_ENV_FVa_CALLS_SUBSET] THENC
         PURE_REWRITE_CONV[EMPTY_ENV_FVa_CALLS_SUBSET] THENC
         DEEP_IMP_BND_NOT_MATCH_CONV THENC
         ONCE_DEPTH_CONV IMP_BND_MATCH_CONV THENC
         string_BND_CONV THENC
         PURE_REWRITE_CONV[FVa_g0_SUBSET;FVa_NULL_ENV_SUBSET] THENC
         PURE_REWRITE_CONV[SOME_ALESS_REC] THENC
         REWRITE_CONV[aexp_distinct;vexp_distinct] THENC
       debug_CONV THENC
         ONCE_DEPTH_CONV FVa_CONV THENC
         ONCE_DEPTH_CONV FVv_CONV THENC
       debug_CONV THENC
         PURE_REWRITE_CONV[APPEND] THENC
         DEPTH_CONV let_CONV THENC
         ONCE_DEPTH_CONV logicals_CONV THENC
         PURE_REWRITE_CONV[APPEND] THENC
       debug_CONV THENC
         PURE_REWRITE_CONV[DISJOINT_LIST;SUBSET;SET_LIST] THENC
         ONCE_DEPTH_CONV (IN_CONV var_EQ_CONV) THENC
         REWRITE_CONV[] THENC
       debug_CONV THENC
         ONCE_DEPTH_CONV WFxs_CONV THENC
         ONCE_DEPTH_CONV WFx_CONV THENC
       debug_CONV THENC
         REWRITE_CONV[];;

let d = [[
   procedure f(var x;val n);
      global a,b;
      pre   true;
      post  a = ^n * 2;

      x := b;
      a := n + n
   end procedure;

   procedure g(var y;val n);
      global a,b,c;
      pre   true;
      post  a = ^n * 3;
      calls f with n = ^n;
      calls g with y = ^c;
      recurses with n < ^n;

      f(c;n);
      a := a + n
   end procedure
      ]] in
% g % WFd_CONV "WFd ^d (mkenv ^d r0)";;
Run time: 12.5s
Garbage collection time: 0.8s

e(CONV_TAC (check_op_CONV `WFd` 2));;
Run time: 7.1s
Garbage collection time: 0.3s
Intermediate theorems generated: 4

e(CONV_TAC (PURE_REWRITE_CONV[WFd;WFproc_syntax;mkenv]));;
Run time: 42.5s
Garbage collection time: 3.1s
Intermediate theorems generated: 240


e(CONV_TAC (ONCE_DEPTH_CONV WFc_CONV));;
Run time: 53.5s
Garbage collection time: 5.3s
Intermediate theorems generated: 1054

e(CONV_TAC (ONCE_DEPTH_CONV GVc_CONV));;
Run time: 25.3s
Garbage collection time: 1.8s
Intermediate theorems generated: 192

e(CONV_TAC (ONCE_DEPTH_CONV FVc_CONV));;
Run time: 28.4s
Garbage collection time: 3.0s
Intermediate theorems generated: 2323


e(CONV_TAC (PURE_REWRITE_CONV[BND_ENV_FVa_CALLS_SUBSET]));;
Run time: 14.9s
Garbage collection time: 1.3s
Intermediate theorems generated: 88

e(CONV_TAC (PURE_REWRITE_CONV[EMPTY_ENV_FVa_CALLS_SUBSET]));;
Run time: 12.7s
Garbage collection time: 0.9s
Intermediate theorems generated: 58

e(CONV_TAC DEEP_IMP_BND_NOT_MATCH_CONV);;
Run time: 16.0s
Garbage collection time: 2.3s
Intermediate theorems generated: 514

e(CONV_TAC (ONCE_DEPTH_CONV IMP_BND_MATCH_CONV));;
Run time: 11.4s
Garbage collection time: 1.0s
Intermediate theorems generated: 164

e(CONV_TAC string_BND_CONV);;
Run time: 10.0s
Garbage collection time: 0.6s
Intermediate theorems generated: 86

e(CONV_TAC (PURE_REWRITE_CONV[FVa_g0_SUBSET;FVa_NULL_ENV_SUBSET]));;
Run time: 10.3s
Garbage collection time: 0.9s
Intermediate theorems generated: 68

e(CONV_TAC (PURE_REWRITE_CONV[SOME_ALESS_REC]));;
Run time: 9.6s
Garbage collection time: 0.8s
Intermediate theorems generated: 29

e(CONV_TAC (REWRITE_CONV[aexp_distinct;vexp_distinct]));;
Run time: 33.9s
Garbage collection time: 3.7s
Intermediate theorems generated: 3279

e(CONV_TAC (ONCE_DEPTH_CONV FVa_CONV));;
Run time: 9.6s
Garbage collection time: 0.9s
Intermediate theorems generated: 414

e(CONV_TAC (ONCE_DEPTH_CONV FVv_CONV));;
Run time: 7.0s
Garbage collection time: 0.7s
Intermediate theorems generated: 25

e(CONV_TAC (PURE_REWRITE_CONV[APPEND]));;
Run time: 8.8s
Garbage collection time: 0.9s
Intermediate theorems generated: 177

e(CONV_TAC (DEPTH_CONV let_CONV));;
Run time: 7.8s
Garbage collection time: 0.8s
Intermediate theorems generated: 123

e(CONV_TAC (ONCE_DEPTH_CONV logicals_CONV));;
Run time: 10.0s
Garbage collection time: 1.0s
Intermediate theorems generated: 387

e(CONV_TAC (PURE_REWRITE_CONV[APPEND]));;
Run time: 9.6s
Garbage collection time: 1.1s
Intermediate theorems generated: 130

e(CONV_TAC (PURE_REWRITE_CONV[DISJOINT_LIST;SUBSET;SET_LIST]));;
Run time: 26.9s
Garbage collection time: 2.2s
Intermediate theorems generated: 845

% ??? %
e(CONV_TAC (PURE_ONCE_REWRITE_CONV[IN_UNION]));;
Run time: 10.0s
Garbage collection time: 1.2s
Intermediate theorems generated: 13

e(CONV_TAC (ONCE_DEPTH_CONV (IN_CONV var_EQ_CONV)));;
Run time: 21.9s
Garbage collection time: 2.4s
Intermediate theorems generated: 3826

e(CONV_TAC (REWRITE_CONV[]));;
Run time: 2.2s
Garbage collection time: 0.3s
Intermediate theorems generated: 148

e(CONV_TAC (ONCE_DEPTH_CONV WFxs_CONV));;
Run time: 3.3s
Garbage collection time: 0.5s
Intermediate theorems generated: 304

e(CONV_TAC (ONCE_DEPTH_CONV WFx_CONV));;
Run time: 0.5s
Intermediate theorems generated: 20

% ??? %
e(CONV_TAC (REWRITE_CONV[FVxs_DEF;SET_LIST]));;
Run time: 1.9s
Garbage collection time: 0.6s
Intermediate theorems generated: 135

% ??? %
e(CONV_TAC (DEPTH_CONV (UNION_CONV var_EQ_CONV)));;
Run time: 0.9s
Intermediate theorems generated: 4

e(CONV_TAC (REWRITE_CONV[]));;
Run time: 2.1s
Intermediate theorems generated: 37
>%



let WFp_CONV =
         REWR_CONV WFp THENC
         RAND_CONV (PURE_REWRITE_CONV[mkenv]) THENC
         let_CONV THENC
         RATOR_CONV (RAND_CONV WFd_CONV) THENC
         RAND_CONV WFc_CONV THENC
         REWRITE_CONV[];;

%<
let pi = [[
program
   procedure f(var x;val n);
      global a,b;
      pre   true;
      post  a = ^n * 2;

      x := b;
      a := n + n
   end procedure;

   procedure g(var y;val n);
      global a,b,c;
      pre   true;
      post  a = ^n * 3;
      calls f with n = ^n;
      calls g with y = ^c;
      recurses with n < ^n;

      f(c;n);
      a := a + n
   end procedure;

   g(z;5)
end program ]] in
 g % WFp_CONV % "WFp ^pi";;
e(PURE_ONCE_REWRITE_TAC[WFp]);;
Run time: 9.1s
Garbage collection time: 0.3s
Intermediate theorems generated: 10

e(PURE_REWRITE_TAC[mkenv]);;
Run time: 9.2s
Garbage collection time: 1.5s
Intermediate theorems generated: 56

e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
Run time: 11.2s
Garbage collection time: 0.5s
Intermediate theorems generated: 12

e(CONV_TAC ((RATOR_CONV o RAND_CONV) WFd_CONV));;
Run time: 136.6s
Garbage collection time: 19.5s
Intermediate theorems generated: 10913

e(CONV_TAC (RAND_CONV WFc_CONV));;
Run time: 15.7s
Garbage collection time: 2.5s
Intermediate theorems generated: 686

e(CONV_TAC (REWRITE_CONV[]));;
>%


let v_subst_CONV =
               check_op_CONV `<|v` 2
         THENC CHANGED_CONV
                (FIRST_CONV (map REWR_CONV (CONJUNCTS v_subst))) ;;

let vs_subst_CONV =
               check_op_CONV `<|vs` 2
         THENC CHANGED_CONV
                (FIRST_CONV (map REWR_CONV (CONJUNCTS vs_subst))) ;;

let a_subst_CONV =
         % debug_CONV THENC %
                FIRST_CONV (map REWR_CONV (CONJUNCTS a_subst))
         THENC TRY_CONV
              ( check_op_CONV `LET` 2
         % THENC debug_CONV %
         THENC (RAND_CONV o RAND_CONV o RAND_CONV)
                   (RATOR_CONV (RAND_CONV FVa_CONV)
                    THENC DELETE_CONV var_EQ_CONV)
         % THENC debug_CONV %
         THENC (RAND_CONV o RAND_CONV)
                  (IMAGE_CONV
                      (PURE_REWRITE_CONV[o_THM]
                       THENC var_BND_CONV
                       THENC PURE_REWRITE_CONV[%IDENT_SS_var;%FVv_DEF]
                       THENC DEPTH_CONV (UNION_CONV var_EQ_CONV))
                      var_EQ_CONV)
         % THENC debug_CONV %
         THENC RAND_CONV (PURE_REWRITE_CONV[UNION_SET] THENC
                          DEPTH_CONV (UNION_CONV var_EQ_CONV))
         % THENC debug_CONV %
         THENC let_CONV
         % THENC debug_CONV %
         THENC RAND_CONV (DEPTH_CONV variant_CONV)
         % THENC debug_CONV %
         THENC let_CONV
              ) ;;

%
new_special_symbol `<|a`;;

a_subst_CONV "^([[ {x*2 < y+1} ]])
                <|a (BND (APLUS (AVAR (VAR `y` 0)) (ANUM 6))
                         (VAR `x` 0) IDENT_SS)";;

a_subst_CONV "^([[ {?y. x*2 < y+1} ]])
                <|a (BND (APLUS (AVAR (VAR `y` 0)) (ANUM 6))
                         (VAR `x` 0) IDENT_SS)";;
%

letrec subst_cl_CONV clean t =
   (% (\t. print_string `:`; debug_CONV t) THENC %
    (IDENT_SS_CONV ORELSEC
     (BND_CONV var_EQ_CONV THENC subst_cl_CONV clean) ORELSEC
     (v_subst_CONV THENC subst_cl_CONV clean) ORELSEC
     (vs_subst_CONV THENC subst_cl_CONV clean) ORELSEC
     ((\t. let (asub,[a;ss]) = strip_comb t in
           (if not ((fst o dest_const)asub = `<|a`) then fail;
            ALL_CONV t)) THENC
      (if clean then ALL_CONV
       else RATOR_CONV (RAND_CONV (subst_cl_CONV clean))) THENC
      a_subst_CONV THENC
      subst_cl_CONV true) ORELSEC
     SUB_CONV (subst_cl_CONV clean))) t;;

let subst_CONV = subst_cl_CONV false;;

%
dest_const "$<|a";;
a_subst_CONV  "(AEQ (AVAR (VAR `x` 0)) (AVAR (VAR `y` 0))) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;
subst_CONV    "(AEQ (AVAR (VAR `x` 0)) (AVAR (VAR `y` 0))) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;
subst_CONV    "^([[ {[1,2] << [x,y]} ]]) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;
a_subst_CONV  "(AFORALL (VAR `x` 0)
                    (AEQ (AVAR (VAR `x` 0)) (AVAR (VAR `y` 0)))) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;
subst_CONV
   "(AAND
       (AEQ (AVAR (VAR `x` 0)) (AVAR (VAR `y` 0)))
       (AEQ (AVAR (VAR `z` 0)) (AVAR (VAR `x` 1)))) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;
subst_CONV    "(AFORALL (VAR `x` 0)
                    (AEQ (AVAR (VAR `x` 0)) (AVAR (VAR `y` 0)))) <|a
               (BND (AVAR (VAR `x` 0)) (VAR `y` 0) IDENT_SS)";;

let t1 =
      " (AAND 
         ATRUE
         ((AFORALL
           (VAR `n` 0)
           (AFORALL
            (VAR `a` 0)
            (AIMP
             (AEQ
              (AMULT (ANUM 2) (AVAR(VAR `a` 0)))
              (AMULT (AVAR(VAR `^n` 0)) (APLUS (AVAR(VAR `^n` 0)) (ANUM 1))) )
             (AEQ (AVAR(VAR `a` 0)) (ANUM 10)) )))
        <|a
          (BND
           (AVAR(VAR `n` 0))
           (VAR `^n` 0)
           (BND(AVAR(VAR `a` 0))(VAR `^a` 0)IDENT_SS)))) <|a
        (BND(ANUM 4)(VAR `n` 0)IDENT_SS) ";;

subst_CONV t1;;


|- (AAND 
    ATRUE
    ((AFORALL
      (VAR `n` 0)
      (AFORALL
       (VAR `a` 0)
       (AIMP
        (AEQ
         (AMULT(ANUM 2)(AVAR(VAR `a` 0)))
         (AMULT(AVAR(VAR `^n` 0))(APLUS(AVAR(VAR `^n` 0))(ANUM 1))))
        (AEQ(AVAR(VAR `a` 0))(ANUM 10))))) <|a
     (BND
      (AVAR(VAR `n` 0))
      (VAR `^n` 0)
      (BND(AVAR(VAR `a` 0))(VAR `^a` 0)IDENT_SS)))) <|a
   (BND(ANUM 4)(VAR `n` 0)IDENT_SS) =
   AAND 
   ATRUE
   (AFORALL
    (VAR `n` 1)
    (AFORALL
     (VAR `a` 0)
     (AIMP
      (AEQ
       (AMULT(ANUM 2)(AVAR(VAR `a` 0)))
       (AMULT(ANUM 4)(APLUS(ANUM 4)(ANUM 1))))
      (AEQ(AVAR(VAR `a` 0))(ANUM 10)))))

%

%
e(CONV_TAC (ONCE_DEPTH_CONV (FIRST_CONV (map REWR_CONV (CONJUNCTS vcg1)))));;
%

let FVa_subst_CONV =
    RAND_CONV subst_CONV
    THENC FVa_CONV;;

let CALL_CONV =
   (\tm. let (lt,[f;x]) = strip_comb tm in
         (if not((fst o dest_const)lt = `LET`) then fail;
          if not(length(strip_pair x) = 8) then fail;
          ALL_CONV tm))
         THENC let_CONV
         %THENC debug_CONV%
         THENC RAND_CONV 
                    (RAND_CONV 
                          (RATOR_CONV (RAND_CONV FVa_subst_CONV)
                           THENC RAND_CONV (PURE_REWRITE_CONV[APPEND;SET_LIST])
                           THENC UNION_CONV var_EQ_CONV)
                     THENC variants_CONV)
         THENC let_CONV
         %THENC debug_CONV%
         THENC PURE_REWRITE_CONV[APPEND_INFIX;APPEND]
         THENC let_CONV
         THENC let_CONV
         %THENC debug_CONV%
         THENC PURE_REWRITE_CONV[APPEND]
         THENC let_CONV
         THENC let_CONV
         %THENC debug_CONV%
         THENC RAND_CONV logicals_CONV
         THENC let_CONV
         THENC RAND_CONV logicals_CONV
         THENC let_CONV
         %THENC debug_CONV%
         THENC RAND_CONV 
                    (RAND_CONV FVa_subst_CONV
                     THENC variants_CONV)
         THENC let_CONV
         %THENC debug_CONV%
         THENC (RATOR_CONV o RAND_CONV o RATOR_CONV o RAND_CONV
              o RATOR_CONV o RAND_CONV o RATOR_CONV o RAND_CONV
              o RAND_CONV) (string_BND_CONV
                            THENC PURE_REWRITE_CONV[EMPTY_PROGRESS_ENV_THM;
                                                    NULL_PROGRESS_ENV])
         %THENC debug_CONV%
         THENC PURE_REWRITE_CONV [var_list_IDENT;a_subst_IDENT]
         %THENC debug_CONV%
         THENC PURE_REWRITE_CONV [assign_list;var_list_THM]
         %THENC debug_CONV%
         THENC PURE_REWRITE_CONV [VES_DEF;VES_state_DEF]
         THENC REPEATC (CHANGED_CONV
            (DEPTH_CONV BETA_CONV THENC
             REWRITE_CONV[VE1_DEF;IDENT_SS_var;
                         VES1_DEF;trans_cont_THM]))
         THENC PURE_REWRITE_CONV [BNDsL]
         THENC PURE_REWRITE_CONV [mk_foralls;IDENT_SS]
         %THENC debug_CONV%
         THENC subst_CONV
         %THENC debug_CONV%
         ;;


let assign_CONV =
    %debug_CONV THENC%
          REWR_CONV assign
    THENC (RATOR_CONV o RATOR_CONV o RAND_CONV) (REWR_CONV VE_DEF)
    THENC (RAND_CONV) (REWR_CONV VE_state_DEF)
    THENC REPEATC (CHANGED_CONV
            (DEPTH_CONV BETA_CONV THENC
             REWRITE_CONV[VE1_DEF;IDENT_SS_var;trans_cont_THM]));;

%<
g [[ {true} even(;n-1) {(?b. ^n = 2*b + a) /\ a < 2 /\ n = ^n} /
        procedure odd(;val n);
           global a;
           pre  true;
           post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
           calls even with n < ^n;
           calls odd  with n < ^n;
           recurses with n < ^n;

           if n = 0 then a:=0
           else if n = 1 then even(;n-1)
                         else odd(;n-2)
                fi
           fi
        end procedure;

        procedure even(;val n);
           global a;
           pre  true;
           post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
           calls odd  with n < ^n;
           calls even with n < ^n;
           recurses with n < ^n;

           if n = 0 then a:=1
           else if n = 1 then odd(;n-1)
                         else even(;n-2)
                fi
           fi
        end procedure
      ]];;
e(REWRITE_TAC[mkenv]);;
e(STRIP_TAC);;
e(STRIP_TAC);;
e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vcgc_THM));;
e(EXISTS_TAC "g0");;
e(ASM_REWRITE_TAC[]);;
e(POP_TAC THEN POP_TAC);;
e(CONV_TAC ((RATOR_CONV o RAND_CONV) WFc_CONV));;
>%

%
|- !b a. ab_pre b a = a <|a (AB_state b)
|- !b. AB_state b = SND(AB1 b IDENT_SS)
%

let ab_pre_CONV =
    REWR_CONV ab_pre
    THENC (RAND_CONV) (REWR_CONV AB_state_DEF)
    THENC REPEATC (CHANGED_CONV
            (DEPTH_CONV BETA_CONV THENC
             REWRITE_CONV[VE1_DEF;IDENT_SS_var;VES1_DEF;
                               AB1_DEF;trans_cont_THM]));;

%
|- !b. AB b = FST(AB1 b IDENT_SS)
%

let AB_CONV =
    REWR_CONV AB_DEF
    THENC REPEATC (CHANGED_CONV
            (DEPTH_CONV BETA_CONV THENC
             REWRITE_CONV[VE1_DEF;IDENT_SS_var;VES1_DEF;
                               AB1_DEF;trans_cont_THM]));;

%
    THENC REPEATC (CHANGED_CONV
            (DEPTH_CONV BETA_CONV THENC
             REWRITE_CONV[VE1_DEF;VE_DEF;VE_state_DEF;IDENT_SS_var;
                               VES1_DEF;VES_DEF;VES_state_DEF;BNDsL;
                               AB1_DEF;AB_DEF;AB_state_DEF;trans_cont_THM]));;
%



% We want to construct conversions to treat each of the functions
  vcg1, vcgc, vcgd, and vcg, by calling on the preceeding ones.
%

let [vcg1_SKIP;vcg1_ABORT;vcg1_ASSIGN;vcg1_SEQ;vcg1_IF;vcg1_WHILE;vcg1_CALL] =
    CONJUNCTS vcg1;;

% DECLARE ALL vcg1_*_CONV and vcg1_CONV, mutually recursively: %

letrec

%
vcg1_SKIP = |- !calls q r. vcg1 SKIP calls q r = q,[]
%

  vcg1_SKIP_CONV t =
    (REWR_CONV vcg1_SKIP) t

%
vcg1_ABORT = |- !calls q r. vcg1 ABORT calls q r = AFALSE,[]
%

and

  vcg1_ABORT_CONV t =
    (REWR_CONV vcg1_ABORT) t

%
vcg1_ASSIGN = 
|- !x e calls q r. vcg1(ASSIGN x e)calls q r = q <|a (x := e),[]
%

and

  vcg1_ASSIGN_CONV t =
   (REWR_CONV vcg1_ASSIGN
    THENC (RATOR_CONV o RAND_CONV o RAND_CONV) assign_CONV
    THENC (RATOR_CONV o RAND_CONV) subst_CONV) t

%
vcg1_SEQ = 
|- !c1 c2 calls q r.
    vcg1(SEQ c1 c2)calls q r =
    (let (s,h2) = vcg1 c2 calls q r
     in
      let (p,h1) = vcg1 c1 calls s r in p,h1 & h2)
%

and

  vcg1_SEQ_CONV t =
   (REWR_CONV vcg1_SEQ
    THENC (RAND_CONV) vcg1_CONV
    THENC let_CONV
    THENC (RAND_CONV) vcg1_CONV
    THENC let_CONV
    THENC (RAND_CONV) (PURE_REWRITE_CONV[APPEND_INFIX;APPEND]) ) t

%
vcg1_IF = 
|- !b c1 c2 calls q r.
    vcg1(IF b c1 c2)calls q r =
    (let (r1,h1) = vcg1 c1 calls q r
     in
      let (r2,h2) = vcg1 c2 calls q r
      in
       ACOND(AB b)(ab_pre b r1)(ab_pre b r2),h1 & h2)
%

and

  vcg1_IF_CONV t =
   (REWR_CONV vcg1_IF
    THENC (RAND_CONV) vcg1_CONV
    THENC let_CONV
    THENC (RAND_CONV) vcg1_CONV
    THENC let_CONV
    THENC (RATOR_CONV o RAND_CONV o DEPTH_CONV) AB_CONV
    THENC (RATOR_CONV o RAND_CONV o DEPTH_CONV) ab_pre_CONV
    THENC subst_CONV
    THENC (RAND_CONV) (PURE_REWRITE_CONV[APPEND_INFIX;APPEND]) ) t

%
vcg1_WHILE = 
|- !a pr b c calls q r.
    vcg1(WHILE a pr b c)calls q r =
    (let (v0,v1) = dest_ALESS pr
     in
      let (p,h) = vcg1 c calls(AAND a pr)r
      in
       a,
       [AIMP(AAND(AAND a(AB b))(AEQ v0 v1))(ab_pre b p);
        AIMP(AAND a(ANOT(AB b)))(ab_pre b q)] &
       h)
%

and

  vcg1_WHILE_CONV t =
   (REWR_CONV vcg1_WHILE
    THENC (RAND_CONV) (REWR_CONV dest_ALESS)
    THENC let_CONV
    THENC (RAND_CONV) vcg1_CONV
    THENC let_CONV
    THENC (RAND_CONV o RATOR_CONV o RAND_CONV o DEPTH_CONV) AB_CONV
    THENC (RAND_CONV o RATOR_CONV o RAND_CONV o DEPTH_CONV) ab_pre_CONV
    THENC subst_CONV
    THENC (RAND_CONV) (PURE_REWRITE_CONV[APPEND_INFIX;APPEND]) ) t


%

vcg1_CALL = 
|- !p xs es calls q r.
    vcg1(CALL p xs es)calls q r =
    (let (vars,vals,glbs,pre,post,calls',rec,c) = r p
     in
      let vals' = variants vals((FVa q) UNION (SL(APPEND xs glbs)))
      in
       let u = xs & vals'
       in
        let v = vars & vals
        in
         let x' = u & glbs
         in
          let y = v & glbs
          in
           let x0 = logicals x'
           in
            let y0 = logicals y
            in
             let x0' = variants x0(FVa q)
             in
              (AAND
               ((AAND pre(calls p)) <|a (u // v))
               ((mk_foralls x'(AIMP(post <|a ((u & x0') // (v & y0)))q)) <|a
                (x' // x0'))) <|a
              (vals' :== es),[])

%

and

  vcg1_CALL_CONV t =
   (REWR_CONV vcg1_CALL
    THENC RAND_CONV string_BND_CONV
    THENC CALL_CONV) t

and

  vcg1_CONV t =
    let (v,[c;calls;q;r]) = strip_comb t in
    let p_thm =
    (%debug_CONV THENC%
     (vcg1_SKIP_CONV ORELSEC
      vcg1_ABORT_CONV ORELSEC
      vcg1_ASSIGN_CONV ORELSEC
      vcg1_SEQ_CONV ORELSEC
      vcg1_IF_CONV ORELSEC
      vcg1_WHILE_CONV ORELSEC
      vcg1_CALL_CONV)) t
    in
    let (comma,[p;h]) = (strip_comb o snd o dest_eq o concl) p_thm in
    (if get_flag_value `print_vcg` then
       (print_string `By the `;
        print_term ((fst o strip_comb) c);
        print_string ` rule, we have`;
        print_newline();
        print_string `   `;
        assignable_print_vcg_term ("PCSPEC(^p,^c,^q,^r)");
        print_newline();
        if not (h = "[]:(aexp)list") then
           (print_string `with verification conditions`;
            print_newline();
            print_string `   `;
            assignable_print_vcg_term h;
            print_newline());
        print_newline());
     p_thm);;

% END OF MUTUAL RECURSIVE DEFINITION of vcg1_CONV. %

%
g "a = vcg1 SKIP g0 (AEQ (ANUM 2) (AVAR (VAR `x` 0))) r";;
e(CONV_TAC (RAND_CONV vcg1_SKIP_CONV));;

g "a = vcg1 ABORT g0 (AEQ (ANUM 2) (AVAR (VAR `x` 0))) r";;
e(CONV_TAC (RAND_CONV vcg1_ABORT_CONV));;

g "a = vcg1 ^([[ x := x+1 ]]) g0 (AEQ (ANUM 2) (AVAR (VAR `x` 0))) r";;
e(CONV_TAC (RAND_CONV vcg1_ASSIGN_CONV));;

g "a = vcg1 ^([[ x := x+1; skip; x := y-x ]])
            g0 (AEQ (ANUM 2) (AVAR (VAR `x` 0))) r";;
e(CONV_TAC (RAND_CONV vcg1_SEQ_CONV));;

g "a = vcg1 ^([[ if x = 0 then y := ++x else x := y-x fi ]])
            g0 (AEQ (AVAR (VAR `x` 0)) (ANUM 2)) r";;
e(CONV_TAC (RAND_CONV vcg1_CONV));;
%

%
|- !p c calls q r.
    vcgc p c calls q r =
    (let (a,h) = vcg1 c calls q r in [AIMP p a] & h)
%

let vcgc_CONV t =
   let (v,[p;c;calls;q;r]) = strip_comb t in
   (REWR_CONV vcgc
    %THENC debug_CONV%
    THENC (RAND_CONV) vcg1_CONV
    %THENC debug_CONV%
    THENC let_CONV
    THENC (\t. if get_flag_value `print_vcg` then
                  (print_string `By precondition strengthening, we have`;
                   print_newline();
                   print_string `   `;
                   assignable_print_vcg_term ("PCSPEC(^p,^c,^q,^r)");
                   print_newline();
                   print_string `with additional verification condition`;
                   print_newline();
                   print_string `   `;
                   assignable_print_vcg_term
                                  ((snd o dest_comb o fst o dest_comb o
                                    snd o dest_comb o fst o dest_comb) t);
                   print_newline();
                   print_newline()
                  );
               ALL_CONV t)
    THENC (RAND_CONV) (PURE_REWRITE_CONV[APPEND_INFIX;APPEND])) t ;;


%
vcgd;;
|- (!p vars vals glbs pre post calls rec c r.
     vcgd(PROC p vars vals glbs pre post calls rec c)r =
     (let x = (vars & vals) & glbs
      in
       let x0 = logicals x in vcgc(AAND(mk_eqs x0 x)pre)c calls post r)) /\
   (!d1 d2 r.
     vcgd(DSEQ d1 d2)r =
     (let h1 = vcgd d1 r in let h2 = vcgd d2 r in h1 & h2)) /\
   (!r. vcgd DEMPTY r = [])
%

let [vcgd_PROC;vcgd_DSEQ;vcgd_DEMPTY] = CONJUNCTS vcgd;;

% DECLARE ALL vcgd_*_CONV and vcgd_CONV, mutually recursively: %

letrec

%
vcgd_PROC = 
|- !p vars vals glbs pre post calls rec c r.
    vcgd(PROC p vars vals glbs pre post calls rec c)r =
    (let x = (vars & vals) & glbs
     in
      let x0 = logicals x in vcgc(AAND(mk_eqs x0 x)pre)c calls post r)
%

  vcgd_PROC_CONV t =
    let (v,[d;r]) = strip_comb t in
    let (proc,[p;vars;vals;glbs;pre;post;calls;rec;c]) = strip_comb d in
    (if get_flag_value(`print_vcg`) then
        (print_string `For procedure `;
         print_string ((fst o dest_const) p);
         print_string `,`;
         print_newline();
         print_newline());
    (REWR_CONV vcgd_PROC
     THENC (RAND_CONV) (PURE_REWRITE_CONV[APPEND_INFIX;APPEND])
     %THENC debug_CONV%
     THENC let_CONV
     %THENC debug_CONV%
     THENC (RAND_CONV) (DEPTH_CONV logicals_CONV)
     %THENC debug_CONV%
     THENC let_CONV
     %THENC debug_CONV%
     THENC PURE_REWRITE_CONV[mk_eqs_THM]
     %THENC debug_CONV%
     THENC vcgc_CONV
     %THENC debug_CONV%) t)

%
vcgd_DSEQ = 
|- !d1 d2 r.
    vcgd(DSEQ d1 d2)r =
    (let h1 = vcgd d1 r in let h2 = vcgd d2 r in h1 & h2)
%

and

  vcgd_DSEQ_CONV t =
    (REWR_CONV vcgd_DSEQ
     %THENC debug_CONV%
     THENC (RAND_CONV) vcgd_CONV
     %THENC debug_CONV%
     THENC let_CONV
     %THENC debug_CONV%
     THENC (RAND_CONV) vcgd_CONV
     %THENC debug_CONV%
     THENC let_CONV
     %THENC debug_CONV%) t

%
vcgd_DEMPTY = |- !r. vcgd DEMPTY r = []
%

and

  vcgd_DEMPTY_CONV t =
    (REWR_CONV vcgd_DEMPTY
     %THENC debug_CONV%) t

and

  vcgd_CONV t =
    (%debug_CONV THENC%
     (vcgd_PROC_CONV ORELSEC
      vcgd_DSEQ_CONV) ORELSEC
      vcgd_DEMPTY_CONV) t ;;

% END OF MUTUAL RECURSIVE DEFINITION of vcgd_CONV. %


% GRAPH TRAVERSAL CONVERSIONS: %


let EMPTY_ENV_THM =
 TAC_PROOF
  (([],
    "!p. r0 p = ([],[],[],AFALSE,ATRUE,(\p'. AFALSE),AFALSE,ABORT)"),
   REWRITE_TAC[EMPTY_ENV]
  );;

% EMPTY_ENV_THM
g "!p. r0 p = ([],[],[],AFALSE,ATRUE,(\p'. AFALSE),AFALSE,ABORT)";;
e(REWRITE_TAC[EMPTY_ENV]);;
%


let call_progress1 =
 TAC_PROOF
  (([],
    "!p1 p2 q r.
      call_progress p1 p2 q r =
      (let (vars,vals,glbs,pre,post,calls,rec,c) = r p1 in
       let a = calls p2 in
       ((a = AFALSE) => 
        ATRUE | 
        let x = APPEND(APPEND vars vals)glbs in
        let x0 = logicals x in
        let x0' = variants x0(FVa q) in
        let (vars',vals',glbs',pre',post',calls',rec',c') = r p2 in
        let y = APPEND(APPEND vars' vals')glbs' in
         (mk_foralls y(AIMP(a <|a (x0' // x0))q)) <|a (x // x0')))"),
   REWRITE_TAC[call_progress]
   THEN ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]
   THEN CONV_TAC (DEPTH_CONV let_CONV)
   THEN REPEAT GEN_TAC
   THEN COND_CASES_TAC
   THEN REWRITE_TAC[]
  );;

% call_progress1
g "!p1 p2 q r.
      call_progress p1 p2 q r =
      (let (vars,vals,glbs,pre,post,calls,rec,c) = r p1 in
       let a = calls p2 in
       ((a = AFALSE) => 
        ATRUE | 
        let x = APPEND(APPEND vars vals)glbs in
        let x0 = logicals x in
        let x0' = variants x0(FVa q) in
        let (vars',vals',glbs',pre',post',calls',rec',c') = r p2 in
        let y = APPEND(APPEND vars' vals')glbs' in
         (mk_foralls y(AIMP(a <|a (x0' // x0))q)) <|a (x // x0')))";;
e(REWRITE_TAC[call_progress]);;
e(ONCE_REWRITE_TAC[ISPEC "(r:env) p" ENV_TUPLE]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(REPEAT GEN_TAC);;
e(COND_CASES_TAC);;

 e(REWRITE_TAC[]);;

 e(REWRITE_TAC[]);;
%



%
(implode o tl o butlast o explode) `abcdef`;;
string_of_int 54;;
dest_const "`abc`";;
dest_const "54";;
dest_string "`abc`";;
(implode o drop_last o tl o explode o fst o dest_const)
    "`abcwxyz`";;
%

letrec drop_last lst =
   if tl lst = [] then []
   else (hd lst).(drop_last (tl lst))
   ? failwith `drop_last`;;

let dest_string s =
   let (st,tp) = dest_const s in
   ((implode o drop_last o tl o explode) st,tp);;



let call_progress_CONV t =
      let (cp,[p1;p2;q;r]) = strip_comb t in
      let cp_thm =
        (REWR_CONV call_progress1
         THENC RAND_CONV (string_BND_CONV THENC
                          PURE_REWRITE_CONV[EMPTY_ENV_THM])
%THENC debug_CONV%
         THENC let_CONV
         THENC RAND_CONV string_BND_CONV
         THENC RAND_CONV (PURE_REWRITE_CONV[NULL_PROGRESS_ENV])
%THENC debug_CONV%
         THENC let_CONV
         THENC REWRITE_CONV[]
%THENC debug_CONV%
         THENC ( REWR_CONV (REFL "ATRUE") ORELSEC

                 (REWRITE_CONV[aexp_distinct]
%THENC debug_CONV%
                  THENC RAND_CONV (PURE_REWRITE_CONV[APPEND])
%THENC debug_CONV%
                  THENC let_CONV
                  THENC RAND_CONV logicals_CONV
%THENC debug_CONV%
                  THENC let_CONV
                  THENC (RAND_CONV o RAND_CONV) FVa_CONV
                  THENC RAND_CONV variants_CONV
%THENC debug_CONV%
                  THENC let_CONV
                  THENC RAND_CONV string_BND_CONV
                  THENC RAND_CONV (PURE_REWRITE_CONV[EMPTY_ENV_THM])
%THENC debug_CONV%
                  THENC let_CONV
                  THENC RAND_CONV (PURE_REWRITE_CONV[APPEND])
                  THENC let_CONV
                  THENC PURE_REWRITE_CONV [var_list_IDENT;a_subst_IDENT]
%THENC debug_CONV%
                  THENC PURE_REWRITE_CONV [var_list_THM]
%THENC debug_CONV%
                  THENC PURE_REWRITE_CONV [mk_foralls]
                  THENC subst_CONV
                 )
               )
        ) t
      in
        let p1t = (string_BND_CONV THENC
                   PURE_REWRITE_CONV[EMPTY_ENV_THM]) (mk_comb(r,p1)) in
        let tpl = (strip_pair o snd o dest_eq o concl) p1t in
        let pre = (hd o tl o tl o tl) tpl in
        let a = (snd o dest_eq o concl) cp_thm in
        let p1s = (fst o dest_string) p1 in
        let p2s = (fst o dest_string) p2 in
        (if get_flag_value `print_vcg` then
           (print_string `By the call graph progress from procedure `;
            print_string p1s;
            print_string ` to `;
            print_string p2s;
            print_string `, we have`;
            print_newline();
            print_string `   `;
            assignable_print_vcg_term
                         "PATH_PROGRESS(AAND ^pre ^a,^p1,[],^p2,^q,^r)";
            print_newline();
            print_newline());
         cp_thm);;

%<

let d = [[
   procedure f(var x;val n);
      global a,b;
      pre   true;
      post  a = ^n * 2;
      calls f with true;

      x := b;
      a := n + n
   end procedure;

   procedure g(var y;val n);
      global a,b,c;
      pre   true;
      post  a = ^n * 3;
      calls f with n = ^n;
      calls g with y = ^c;
      recurses with n < ^n;

      f(c;n);
      a := a + n
   end procedure
      ]] in
(REWRITE_CONV[mkenv] THENC call_progress_CONV)
  "call_progress `g` `f` ^([[ {x = 2*^n} ]]) (mkenv ^d r0)";;

By the call graph progress from procedure g to f, we have
   [[ {true /\ !x. !n1. !a. !b. n1 = n ==> x = 2 * ^n}
      g-<>->f
      {x = 2 * ^n} ]]

Run time: 132.8s
Garbage collection time: 16.5s
Intermediate theorems generated: 15640

  "call_progress `f` `g` ^([[ {x = 2*^n} ]]) (mkenv ^d r0)";;

By the call graph progress from procedure f to g, we have
   [[ {true /\ true} f-<>->g {x = 2 * ^n} ]]

Run time: 42.6s
Garbage collection time: 5.8s
Intermediate theorems generated: 3277
>%



let FULL_PROGRESS_ENV =
 TAC_PROOF
  (([],
    "!p:string. (\p. ATRUE) p = ATRUE"),
   BETA_TAC
   THEN REWRITE_TAC[]
  );;

% FULL_PROGRESS_ENV
g "!p:string. (\p. ATRUE) p = ATRUE";;
e(BETA_TAC);;
e(REWRITE_TAC[]);;
%


% DECLARE extend_graph_vcs_CONV & fan_out_graph_vcs_CONV, mutually recursively: %

letrec

%
extend_graph_vcs =
|- !p ps p0 q pcs r all_ps n p'.
    extend_graph_vcs p ps p0 q pcs r all_ps n p' =
    (let q1 = call_progress p' p q r
     in
      ((q1 = ATRUE) => 
       [] | 
       ((p' = p0) => 
        let (vars,vals,glbs,pre,post,calls,rec,c) = r p0
        in
         [AIMP(AAND pre(induct_pre rec))q1] | 
        (p' IN (SL(CONS p ps)) => 
         [AIMP(pcs p')q1] | 
         fan_out_graph_vcs p'(CONS p ps)p0 q1(BND q1 p' pcs)r all_ps n))))
%

  extend_graph_vcs_CONV t =
        (REWR_CONV extend_graph_vcs
%THENC debug_CONV%
         THENC RAND_CONV call_progress_CONV
         THENC let_CONV
         THENC (RATOR_CONV o RATOR_CONV o RAND_CONV)
                            (REWRITE_CONV[]
                             THENC REWRITE_CONV[aexp_distinct])
         THENC COND_CONV
%THENC debug_CONV%
         THENC ( REWR_CONV (REFL "[]:aexp list") ORELSEC

                 (((RATOR_CONV o RATOR_CONV o RAND_CONV) string_EQ_CONV
                   THENC COND_CONV)
%THENC debug_CONV%
         THENC ( (REWR_CONV (REFL "LET f ((r:env) p0) :aexp list")
                  THENC (RAND_CONV) (string_BND_CONV THENC
                                     PURE_REWRITE_CONV[EMPTY_ENV_THM])
                  THENC let_CONV
                  THENC PURE_REWRITE_CONV[induct_pre]
                  THENC (\tm.
     let (c,[a;n]) = strip_comb tm in
     (if get_flag_value `print_vcg` then
      print_string `Generating the undiverted recursion verification condition`;
      print_newline();
      print_string `   `;
      assignable_print_vcg_term a;
      print_newline();
      print_newline());
                          ALL_CONV tm)
                 )

                 ORELSEC

                 ((RATOR_CONV o RATOR_CONV o RAND_CONV)
                                 (REWRITE_CONV[SET_LIST]
                                  THENC (IN_CONV string_EQ_CONV))
                  THENC COND_CONV
         THENC ( (REWR_CONV (REFL "[a:aexp]")
                  THENC (RATOR_CONV o RAND_CONV)
                                 (string_BND_CONV
                                  THENC REWRITE_CONV[FULL_PROGRESS_ENV])
                  THENC (\tm.
     let (c,[a;n]) = strip_comb tm in
     (if get_flag_value `print_vcg` then
      print_string `Generating the diversion verification condition`;
      print_newline();
      print_string `   `;
      assignable_print_vcg_term a;
      print_newline();
      print_newline());
                          ALL_CONV tm)
                 )

                 ORELSEC

                 (fan_out_graph_vcs_CONV
                 )
               )))))) t

%
fan_out_graph_vcs = 
|- !p ps p0 q r all_ps n.
    fan_out_graph_vcs p ps p0 q pcs r all_ps(SUC n) =
    FLAT(MAP(extend_graph_vcs p ps p0 q pcs r all_ps n)all_ps)
%

and

  fan_out_graph_vcs_CONV t =
    (REWR_CONV fan_out_graph_vcs
%THENC debug_CONV%
     THENC (RAND_CONV) (MAP_CONV extend_graph_vcs_CONV)
%THENC debug_CONV%
     THENC FLAT_CONV
%THENC debug_CONV%
     ) t ;;

% END OF MUTUAL RECURSIVE DEFINITION of graph traversal conversions. %


%<
(REWRITE_CONV[mkenv] THENC
 fan_out_graph_vcs_CONV)
    "fan_out_graph_vcs `odd` [] `odd` ^([[ {n < ^n} ]]) (\p.ATRUE)
        (mkenv
          ^([[ procedure odd (var a; val n);
                  pre  true;
                  post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls odd  with n < ^n;
                  calls even with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 0
                  else if n = 1
                     then even(a;n - 1)
                     else odd (a;n - 2)
                  fi fi
               end procedure;

               procedure even (var a; val n);
                  pre  true;
                  post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls even with n < ^n;
                  calls odd  with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 1
                  else if n = 1
                     then odd (a;n - 1)
                     else even(a;n - 2)
                  fi fi
               end procedure
            ]])
          r0)
        [`odd`;`even`] (SUC (SUC 0))";;
>%


%
graph_vcs =
|- !all_ps r p.
    graph_vcs all_ps r p =
    (let (vars,vals,glbs,pre,post,calls,rec,c) = r p
     in
      fan_out_graph_vcs p[]p rec(\p'. ATRUE)r all_ps(LENGTH all_ps))
%

let graph_vcs_CONV =
    (\t. 
      let (g,[a;r;p]) = strip_comb t in
         if get_flag_value(`print_vcg`) then
            (print_string `Traversing the call graph back from the procedure `;
             print_string ((fst o dest_string) p);
             print_string `:`;
             print_newline();
             print_newline());
         ALL_CONV t)
    THENC REWR_CONV graph_vcs
    %THENC debug_CONV%
    THENC (RAND_CONV) (string_BND_CONV THENC
                       PURE_REWRITE_CONV[EMPTY_ENV_THM])
    %THENC debug_CONV%
    THENC let_CONV
    %THENC debug_CONV%
    THENC (RAND_CONV) (PURE_REWRITE_CONV[LENGTH])
    THENC fan_out_graph_vcs_CONV
    %THENC debug_CONV%;;

%<
(REWRITE_CONV[mkenv] THENC
 graph_vcs_CONV)
    "graph_vcs [`odd`;`even`]
        (mkenv
          ^([[ procedure odd (var a; val n);
                  pre  true;
                  post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls odd  with n < ^n;
                  calls even with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 0
                  else if n = 1
                     then even(a;n - 1)
                     else odd (a;n - 2)
                  fi fi
               end procedure;

               procedure even (var a; val n);
                  pre  true;
                  post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls even with n < ^n;
                  calls odd  with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 1
                  else if n = 1
                     then odd (a;n - 1)
                     else even(a;n - 2)
                  fi fi
               end procedure
            ]])
          r0)
         `odd`";;
>%


%
vcgg =
|- !all_ps r. vcgg all_ps r = FLAT(MAP(graph_vcs all_ps r)all_ps)
%

let vcgg_CONV =
    (\t. 
      let (v,[a;r]) = strip_comb t in
         if get_flag_value(`print_vcg`) then
            (print_string `Examining the structure of the procedure call graph:`;
             print_newline();
             print_newline());
         ALL_CONV t)
    THENC REWR_CONV vcgg
    %THENC debug_CONV%
    THENC (RAND_CONV) (MAP_CONV graph_vcs_CONV)
    %THENC debug_CONV%
    THENC FLAT_CONV
    %THENC debug_CONV%;;

%<
(REWRITE_CONV[mkenv] THENC
 vcgg_CONV)
    "vcgg [`odd`;`even`]
        (mkenv
          ^([[ procedure odd (var a; val n);
                  pre  true;
                  post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls odd  with n < ^n;
                  calls even with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 0
                  else if n = 1
                     then even(a;n - 1)
                     else odd (a;n - 2)
                  fi fi
               end procedure;

               procedure even (var a; val n);
                  pre  true;
                  post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls even with n < ^n;
                  calls odd  with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 1
                  else if n = 1
                     then odd (a;n - 1)
                     else even(a;n - 2)
                  fi fi
               end procedure
            ]])
          r0)
         ";;
>%


%
vcg;;
|- !d c q.
    vcg(PROG d c)q =
    (let r = mkenv d r0
     in
      let h1 = vcgd d r
      in
       let h2 = vcgg(proc_names d)r
       in
        let h3 = vcgc ATRUE c g0 q r in h1 & (h2 & h3))
%

let vcg_CONV =
    REWR_CONV vcg
    %THENC debug_CONV%
    THENC (RAND_CONV) (PURE_REWRITE_CONV[mkenv])
    %THENC debug_CONV%
    THENC let_CONV
    %THENC debug_CONV%
    THENC (RAND_CONV) vcgd_CONV
    %THENC debug_CONV%
    THENC let_CONV
    %THENC debug_CONV%
    THENC (RAND_CONV o RATOR_CONV o RAND_CONV)
                        (PURE_REWRITE_CONV[proc_names]
                         THENC PURE_REWRITE_CONV[APPEND_INFIX;APPEND])
    THENC (RAND_CONV) vcgg_CONV
    %THENC debug_CONV%
    THENC let_CONV
    THENC
    (\t. if get_flag_value(`print_vcg`) then
            (print_string `For the main body,`;
             print_newline();
             print_newline());
         ALL_CONV t)
    THENC (RAND_CONV) vcgc_CONV
    %THENC debug_CONV%
    THENC let_CONV
    THENC (PURE_REWRITE_CONV[APPEND_INFIX;APPEND])
    %THENC debug_CONV%;;


%<
(vcg_CONV)
    "vcg
       ^([[ program
               procedure odd (var a; val n);
                  pre  true;
                  post (?b.^n = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls odd  with n < ^n;
                  calls even with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 0
                  else if n = 1
                     then even(a;n - 1)
                     else odd (a;n - 2)
                  fi fi
               end procedure;

               procedure even (var a; val n);
                  pre  true;
                  post (?b.^n + 1 = 2*b + a) /\ a < 2 /\ n = ^n;
                  calls even with n < ^n;
                  calls odd  with n < ^n;
                  recurses   with n < ^n;
                  
                  if n = 0
                     then a := 1
                  else if n = 1
                     then odd (a;n - 1)
                     else even(a;n - 2)
                  fi fi
               end procedure;

               odd(a;5)
            end program
            ]])
          ^([[ {a = 1} ]])
         ";;
>%



let var_name var = 
   let (vr,[basetm;indextm]) = strip_comb var in
   let base = (fst o dest_string) basetm in
   let indexstr = (fst o dest_const) indextm in
   let index = int_of_string indexstr in
   if index = 0 then base
                else base ^ indexstr;;


let INTERPRET_aexp_CONV =
  (\tm. let (op,[a;s]) = strip_comb tm in
        (if not(op = "A") then fail;
         let (aop,args) = strip_comb a in
         (CHANGED_CONV
                (FIRST_CONV (map REWR_CONV (CONJUNCTS A_DEF)))
          THENC
          (if ((aop = "AFORALL") or (aop = "AEXISTS")) then
            ONCE_DEPTH_CONV (ALPHA_CONV (mk_var (var_name (hd args),":num")))
           else ALL_CONV)) tm));;


letrec collect_prog_vars tm =
    if is_comb tm then
       (let rator,rands = strip_comb tm in
         if rator = "VAR" then [hd rands,hd(tl rands)]
         else flat (map collect_prog_vars (rator . rands)))
    if is_abs tm then
       (let bv,body = dest_abs tm in
        collect_prog_vars body)
    else [];;

let INTERPRET_PROG_VARS_TAC :tactic = \ (asl,gl).
     (let vars = (setify o rev o collect_prog_vars) gl in
      let names = map (\ (basetm,indextm).
                        let base = (fst o dest_string) basetm in
                        let index = (int_of_string o fst o dest_const) indextm
                        in
                        if index = 0 then base
                        else base ^ (string_of_int index)) vars in
      let spec_tactics =
           map (\ ((base,index),name).
                     SPEC_TAC("(s:state)(VAR ^base ^index)",
                              mk_var(name,":num")))
               (combine(vars,names)) in
      EVERY spec_tactics (asl,gl) )
      ? failwith `INTERPRET_PROG_VARS_TAC`;;

%<
let MULT_CONV1 =
  (\tm. let (m,[x;y]) = strip_comb tm in
        (if not (m = "$*") then fail;
         %let xs = (int_of_string o fst o dest_const) x in
         let ys = (int_of_string o fst o dest_const) y in%
         if x = "0" then ALL_CONV tm
         else AP_THM (AP_TERM m (num_CONV x)) y))
  THENC PURE_REWRITE_CONV[MULT];;

let MULT_CONV =
  MULT_CONV1
  THENC REPEATC (CHANGED_CONV (ONCE_DEPTH_CONV MULT_CONV1))
  THENC REPEATC (CHANGED_CONV (ONCE_DEPTH_CONV ADD_CONV))
  THENC PURE_REWRITE_CONV[ADD];;

MULT_CONV1 "4 * 5";;
MULT_CONV1 "0 * 5";;
MULT_CONV  "4 * 5";;
MULT_CONV  "0 * 5";;
>%


let VCGC_TAC =
    REWRITE_TAC[mkenv]
    THEN DISCH_TAC
    THEN MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vcgc_THM)
    THEN POP_ASSUM (\th. REWRITE_TAC[th])
    THEN CONV_TAC (DEPTH_CONV WFc_CONV)
    THEN CONV_TAC (DEPTH_CONV vcgc_CONV)
    THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL;CLOSE]
    THEN CONV_TAC (TOP_DEPTH_CONV INTERPRET_aexp_CONV)
    THEN REWRITE_TAC[VS_DEF;V_DEF]
    THEN CONV_TAC var_BND_CONV
    THEN REPEAT CONJ_TAC
    THEN ( GEN_TAC ORELSE ALL_TAC )
    THEN INTERPRET_PROG_VARS_TAC;;

%< VCGC_TAC
e(DISCH_TAC);;
e(MATCH_MP_TAC (REWRITE_RULE[AND_IMP_INTRO] vcgc_THM));;

% The following shows the environment is well formed: %

e(POP_ASSUM (\th. REWRITE_TAC[th]));;

% The following reduces the goal of WFc c r: %

e(CONV_TAC (DEPTH_CONV WFc_CONV));;

let st2 = get_state();;
set_state(st2);;

% The following 'runs' the VCGC function. %

e(CONV_TAC (DEPTH_CONV vcgc_CONV));;

% The following distributes the CLOSE operator over each VC, and
  interprets each VC into the HOL logic. %

e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL;CLOSE]);;
e(CONV_TAC (TOP_DEPTH_CONV INTERPRET_aexp_CONV));;
e(REWRITE_TAC[VS_DEF;V_DEF]);;
e(CONV_TAC var_BND_CONV);;
e(REPEAT CONJ_TAC
    THEN ( GEN_TAC ORELSE ALL_TAC )
    THEN INTERPRET_PROG_VARS_TAC);;

% Now the specific solving within HOL can begin. %
>%


let VCG_TAC =
    MATCH_MP_TAC vcg_THM
    THEN CONV_TAC (DEPTH_CONV WFp_CONV)
    THEN CONV_TAC (DEPTH_CONV vcg_CONV)
    THEN REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL;CLOSE]
    THEN CONV_TAC (TOP_DEPTH_CONV INTERPRET_aexp_CONV)
    THEN REWRITE_TAC[VS_DEF;V_DEF]
    THEN CONV_TAC var_BND_CONV
    THEN REPEAT CONJ_TAC
    THEN ( GEN_TAC ORELSE ALL_TAC )
    THEN INTERPRET_PROG_VARS_TAC;;

%<

e(MATCH_MP_TAC vcg_THM);;
e(CONV_TAC (DEPTH_CONV WFp_CONV));;
e(CONV_TAC (DEPTH_CONV vcg_CONV));;
e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL;CLOSE]);;
e(CONV_TAC (TOP_DEPTH_CONV INTERPRET_aexp_CONV));;
e(REWRITE_TAC[VS_DEF;V_DEF]);;
e(CONV_TAC var_BND_CONV);;
e(REPEAT CONJ_TAC
  THEN ( GEN_TAC ORELSE ALL_TAC )
  THEN INTERPRET_PROG_VARS_TAC);;

>%


%<
e(MATCH_MP_TAC vcg_THM);;
e(PURE_REWRITE_TAC[mkenv]);;
e(CONV_TAC (ONCE_DEPTH_CONV WFp_CONV));;

   e(PURE_REWRITE_TAC[WFp]);;
   e(CONV_TAC (ONCE_DEPTH_CONV let_CONV));;
   e(CONV_TAC (ONCE_DEPTH_CONV WFd_CONV));;
   e(CONV_TAC (ONCE_DEPTH_CONV WFc_CONV));;

% This all handled the need to prove WFp pi r. %
% Now for the real vcg function on programs! %

e(REWRITE_TAC[vcg;vcgd]);;
e(REWRITE_TAC[mkenv;APPEND_INFIX;APPEND]);;
e(CONV_TAC (DEPTH_CONV let_CONV));;
e(CONV_TAC (DEPTH_CONV logicals_CONV));;
e(REWRITE_TAC[mk_eqs_THM]);;
let st1 = get_state();;

% The following 'runs' the vcgc function. %

e(REWRITE_TAC[vcgc;vcg1]);;

% The following reduces all references to procedures in the
  current environment, extracting the correct tuple.        %

e(CONV_TAC string_BND_CONV);;

% The following 'runs' the ':=', ':==', and 'ab_pre' operators to create 
  substitutions, which invoke the translation operators. %

e(REWRITE_TAC[assign;assign_list;ab_pre;var_list_THM]);;

% The following 'runs' the translation operators, to reduce the
  substitutions being created to their base forms. %

e(REPEAT (CHANGED_TAC
            (BETA_TAC THEN
             REWRITE_TAC[VE1_DEF;VE_DEF;VE_state_DEF;IDENT_SS_var;
                         VES1_DEF;VES_DEF;VES_state_DEF;BNDsL;
                         AB1_DEF;AB_DEF;AB_state_DEF;trans_cont_THM])));;
e(REWRITE_TAC[a_subst_IDENT]);;

% The following 'executes' the CALL's, '<|a's, and 'let's, bottom up. %

e(CONV_TAC vcg_CONV);;

% The following distributes the CLOSE operator over each VC, and
  interprets each VC into the HOL logic. %

e(REWRITE_TAC[APPEND_INFIX;APPEND;ALL_EL;CLOSE]);;
e(CONV_TAC (TOP_DEPTH_CONV INTERPRET_aexp_CONV));;
e(REWRITE_TAC[VS_DEF;V_DEF]);;
e(CONV_TAC var_BND_CONV);;
e(REPEAT CONJ_TAC
    THEN ( GEN_TAC ORELSE ALL_TAC )
    THEN INTERPRET_PROG_VARS_TAC);;
>%
